from e2eml.model_utils.tabular_gan_for_classification import (
    TabularGeneratorClassification,
)
from e2eml.model_utils.tabular_gan_for_regression import TabularGeneratorRegression


class TabularGan(TabularGeneratorClassification, TabularGeneratorRegression):
    def train_tabular_gans(self):
        if self.prediction_mode:
            pass
        else:
            self.set_random_seed()
            if self.class_problem in ("binary", "multiclass"):
                self.train_class_generators()
                self.create_synthetic_data()
            elif self.class_problem == "regression":
                self.train_regression_generators()
                self.create_synthetic_data_regression()
            else:
                pass



import gc
import logging
import os

import numpy as np
import pandas as pd
import psutil
import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd.variable import Variable
from torch.distributions.studentT import StudentT
from torch.utils.data import DataLoader, Dataset

from e2eml.full_processing.postprocessing import FullPipeline

# from tqdm import tqdm
# from transformers import AdamW, get_linear_schedule_with_warmup


# specify GPU
scaler = torch.cuda.amp.GradScaler()  # GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class GanDataset(Dataset):
    def __init__(self, X_data, y_data):
        self.X_data = X_data
        self.y_data = y_data

    def __getitem__(self, index):
        return self.X_data[index], self.y_data[index]

    def __len__(self):
        return len(self.X_data)


class TabularGeneratorRegression(FullPipeline, GanDataset):
    def get_generator_discriminator_regression(self):
        class GeneratorRegression(nn.Module):
            """
            Creates fake data trying to trick the Discriminator.
            """

            def __init__(
                self,
                num_features,
                output_dim=1,
                sign_size=32,
                cha_input=16,
                cha_hidden=32,
                K=2,
                dropout_input=0.2,
                dropout_hidden=0.2,
                dropout_output=0.2,
            ):
                super().__init__()

                hidden_size = sign_size * cha_input
                sign_size1 = sign_size
                sign_size2 = sign_size // 2
                output_size = (sign_size // 4) * cha_hidden

                self.hidden_size = hidden_size
                self.cha_input = cha_input
                self.cha_hidden = cha_hidden
                self.K = K
                self.sign_size1 = sign_size1
                self.sign_size2 = sign_size2
                self.output_size = output_size
                self.dropout_input = dropout_input
                self.dropout_hidden = dropout_hidden
                self.dropout_output = dropout_output

                self.batch_norm1 = nn.BatchNorm1d(num_features)
                self.dropout1 = nn.Dropout(dropout_input)
                dense1 = nn.Linear(num_features, hidden_size, bias=False)
                self.dense1 = nn.utils.weight_norm(dense1)

                # 1st conv layer
                self.batch_norm_c1 = nn.BatchNorm1d(cha_input)
                conv1 = nn.Conv1d(
                    cha_input,
                    cha_input * K,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_input,
                    bias=False,
                )
                self.conv1 = nn.utils.weight_norm(conv1, dim=None)

                self.ave_po_c1 = nn.AdaptiveAvgPool1d(output_size=sign_size2)

                # 2nd conv layer
                self.batch_norm_c2 = nn.BatchNorm1d(cha_input * K)
                self.dropout_c2 = nn.Dropout(dropout_hidden)
                conv2 = nn.Conv1d(
                    cha_input * K,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv2 = nn.utils.weight_norm(conv2, dim=None)

                # 3rd conv layer
                self.batch_norm_c3 = nn.BatchNorm1d(cha_hidden)
                self.dropout_c3 = nn.Dropout(dropout_hidden)
                conv3 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv3 = nn.utils.weight_norm(conv3, dim=None)

                # 4th conv layer
                self.batch_norm_c4 = nn.BatchNorm1d(cha_hidden)
                conv4 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_hidden,
                    bias=False,
                )
                self.conv4 = nn.utils.weight_norm(conv4, dim=None)

                self.avg_po_c4 = nn.AvgPool1d(kernel_size=4, stride=2, padding=1)

                self.flt = nn.Flatten()

                self.batch_norm2 = nn.BatchNorm1d(output_size)
                self.dropout2 = nn.Dropout(dropout_output)
                dense2 = nn.Linear(output_size, output_dim, bias=False)
                self.dense2 = dense2

                # self.loss = nn.BCEWithLogitsLoss()

            def forward(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                return x

            def predict(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                return x

        class DiscriminatorRegression(nn.Module):
            """
            Will decide, if data is real or fake.
            """

            def __init__(
                self,
                num_features,
                output_dim=1,
                sign_size=32,
                cha_input=16,
                cha_hidden=32,
                K=2,
                dropout_input=0.2,
                dropout_hidden=0.2,
                dropout_output=0.2,
            ):
                super().__init__()

                hidden_size = sign_size * cha_input
                sign_size1 = sign_size
                sign_size2 = sign_size // 2
                output_size = (sign_size // 4) * cha_hidden

                self.hidden_size = hidden_size
                self.cha_input = cha_input
                self.cha_hidden = cha_hidden
                self.K = K
                self.sign_size1 = sign_size1
                self.sign_size2 = sign_size2
                self.output_size = output_size
                self.dropout_input = dropout_input
                self.dropout_hidden = dropout_hidden
                self.dropout_output = dropout_output

                self.batch_norm1 = nn.BatchNorm1d(num_features)
                self.dropout1 = nn.Dropout(dropout_input)
                dense1 = nn.Linear(num_features, hidden_size, bias=False)
                self.dense1 = nn.utils.weight_norm(dense1)

                # 1st conv layer
                self.batch_norm_c1 = nn.BatchNorm1d(cha_input)
                conv1 = nn.Conv1d(
                    cha_input,
                    cha_input * K,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_input,
                    bias=False,
                )
                self.conv1 = nn.utils.weight_norm(conv1, dim=None)

                self.ave_po_c1 = nn.AdaptiveAvgPool1d(output_size=sign_size2)

                # 2nd conv layer
                self.batch_norm_c2 = nn.BatchNorm1d(cha_input * K)
                self.dropout_c2 = nn.Dropout(dropout_hidden)
                conv2 = nn.Conv1d(
                    cha_input * K,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv2 = nn.utils.weight_norm(conv2, dim=None)

                # 3rd conv layer
                self.batch_norm_c3 = nn.BatchNorm1d(cha_hidden)
                self.dropout_c3 = nn.Dropout(dropout_hidden)
                conv3 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv3 = nn.utils.weight_norm(conv3, dim=None)

                # 4th conv layer
                self.batch_norm_c4 = nn.BatchNorm1d(cha_hidden)
                conv4 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_hidden,
                    bias=False,
                )
                self.conv4 = nn.utils.weight_norm(conv4, dim=None)

                self.avg_po_c4 = nn.AvgPool1d(kernel_size=4, stride=2, padding=1)

                self.flt = nn.Flatten()

                self.batch_norm2 = nn.BatchNorm1d(output_size)
                self.dropout2 = nn.Dropout(dropout_output)
                dense2 = nn.Linear(output_size, output_dim, bias=False)
                self.dense2 = dense2

                # self.loss = nn.BCEWithLogitsLoss()

            def forward(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                return x

            def predict(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                return x

        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        X_train[self.target_variable] = Y_train
        n_features = X_train.values.shape[1]
        generator = GeneratorRegression(num_features=n_features, output_dim=n_features)
        discriminator = DiscriminatorRegression(num_features=n_features)

        generator.to(device)
        discriminator.to(device)
        return generator, discriminator

    def noise(self, n=None):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        X_train[self.target_variable] = Y_train
        n_features = X_train.values.shape[1]
        if not n:
            n = X_train.values.shape[0]
        m = StudentT(1)
        return m.sample([n, n_features]).to(device)

    def make_ones(self, size):
        data = Variable(torch.ones(size, 1))
        return data.to(device)

    def make_zeros(self, size):
        data = Variable(torch.zeros(size, 1))
        return data.to(device)

    def gan_model_setup_regression(self):
        generator, discriminator = self.get_generator_discriminator_regression()
        # Initialize weights

        g_optim = optim.Adam(
            generator.parameters(),
            lr=self.gan_settings["generator_learning_rate"],
            betas=(0.5, 0.999),
        )
        d_optim = optim.Adam(
            discriminator.parameters(),
            lr=self.gan_settings["discriminator_learning_rate"],
            betas=(0.5, 0.999),
        )

        loss_fn = nn.BCEWithLogitsLoss()
        return generator, discriminator, g_optim, d_optim, loss_fn

    def train_discriminator_regression(
        self, optimizer, real_data, fake_data, discriminator, loss_fn
    ):
        n = real_data.size(0)
        optimizer.zero_grad()
        prediction_real = discriminator(real_data)

        prediction_fake = discriminator(fake_data)

        real_loss = loss_fn(prediction_real - prediction_fake, self.make_ones(n))
        fake_loss = loss_fn(prediction_fake - prediction_real, self.make_zeros(n))
        D_loss = (real_loss + fake_loss) / 2

        D_loss.backward()
        optimizer.step()

        return D_loss

    def train_generator_regression(
        self, optimizer, real_data, fake_data, discriminator, loss_fn
    ):
        n = real_data.size(0)
        optimizer.zero_grad()

        real_pred = discriminator(real_data).detach()
        fake_pred = discriminator(fake_data)

        G_loss = loss_fn(fake_pred - real_pred, self.make_ones(n))

        G_loss.backward()
        optimizer.step()

        return G_loss, optimizer

    def create_gan_train_dataset_regression(self):
        logging.info("Create NLP train dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        X_train[self.target_variable] = Y_train
        self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

        train_dataset = GanDataset(
            torch.from_numpy(X_train.to_numpy().astype(float)).float(),
            torch.from_numpy(Y_train.to_numpy().astype(float)).float(),
        )
        return train_dataset

    def create_gan_train_dataloader_regression(
        self, train_batch_size=None, workers=None
    ):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.gan_settings["batch_size"]

        if workers:
            pass
        else:
            workers = self.gan_settings["num_workers"]
        logging.info("Create Neural network train dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        train_dataset = self.create_gan_train_dataset_regression()
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            num_workers=workers,
            pin_memory=True,
        )
        return train_dataloader

    def create_gan_test_dataloader_regression(
        self, train_batch_size=None, workers=None
    ):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.gan_settings["batch_size"]

        if workers:
            pass
        else:
            workers = self.gan_settings["num_workers"]
        logging.info("Create Neural network train dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        train_dataset = self.create_gan_test_dataset_regression()
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            num_workers=workers,
            pin_memory=True,
        )
        return train_dataloader

    def train_gan_regression(self):
        epochs = self.gan_settings["max_epochs"]
        k = self.gan_settings["discriminator_extra_training_rounds"]
        best_epoch = -1
        # test_noise = self.noise(64)
        train_dataloader = self.create_gan_train_dataloader_regression()
        (
            generator,
            discriminator,
            g_optim,
            d_optim,
            loss_fn,
        ) = self.gan_model_setup_regression()

        generator.train()
        discriminator.train()

        g_losses = []
        d_losses = []
        best_loss = 1000000000000

        for epoch in range(epochs):
            g_loss = 0.0
            d_loss = 0.0
            for i, data in enumerate(train_dataloader):
                imgs, _ = data
                n = len(imgs)
                for _j in range(k):
                    fake_data = generator(self.noise(n)).detach()
                    real_data = imgs.to(device)
                    d_loss += self.train_discriminator_regression(
                        d_optim,
                        real_data,
                        fake_data,
                        discriminator=discriminator,
                        loss_fn=loss_fn,
                    )
                fake_data = generator(self.noise(n))
                real_data = imgs.to(device)
                g_loss_it, g_optim = self.train_generator_regression(
                    g_optim,
                    real_data,
                    fake_data,
                    discriminator=discriminator,
                    loss_fn=loss_fn,
                )
                g_loss = g_loss + g_loss_it

                del fake_data
                del real_data
                _ = gc.collect()

                # img = generator(test_noise).cpu().detach()
                g_losses.append(g_loss / i)
                d_losses.append(d_loss / i)

            if epoch == 0 or g_loss < best_loss:
                print(
                    f"Found better model in epoch {epoch} with generator loss {g_loss}"
                    f" and discriminator loss {d_loss}."
                )
                best_epoch = epoch
                best_loss = g_loss
                state = {
                    "state_dict": generator.state_dict(),
                    "optimizer_dict": g_optim.state_dict(),
                    "bestscore": g_loss,
                }
                torch.save(state, "generator_model.pth")

            if epoch > best_epoch + self.gan_settings["early_stopping_rounds"]:
                break

                # clear_output()
        del discriminator
        del generator
        torch.cuda.empty_cache()
        _ = gc.collect()

    def train_regression_generators(self):
        self.get_current_timestamp(task="Start training GANs.")
        self.train_gan_regression()
        self.get_current_timestamp(task="Finished training GANs.")

    def load_generator_model_states_regression(self, path=None):
        logging.info("Load model save states.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if path:
            pass
        else:
            path = os.getcwd()
        if self.prediction_mode:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if "generator_model.pth" in s
            ]
            return pthes
        else:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if "generator_model.pth" in s
            ]
            return pthes

    def create_synthetic_data_regression(self):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        columns = X_train.columns

        rows_to_create = self.gan_settings["nb_synthetic_rows_to_create"]
        batch_size = self.gan_settings["batch_size"]
        nb_batches = int(rows_to_create / batch_size)
        new_data = []
        (
            generator,
            discriminator,
            g_optim,
            d_optim,
            loss_fn,
        ) = self.gan_model_setup_regression()
        pathes = self.load_generator_model_states_regression(
            path=self.gan_model_save_states_path
        )
        print(pathes)
        for m_path in pathes:
            print(f"Load {m_path}")
            state = torch.load(m_path)
            generator.load_state_dict(state["state_dict"])
            generator.to(device)
            generator.eval()

            with torch.no_grad():
                for _batch in range(nb_batches):
                    synth_data = generator.predict(self.noise(batch_size))
                    synth_data = synth_data.cpu().detach()
                    new_data.append(synth_data)

            del state
            torch.cuda.empty_cache()
            _ = gc.collect()

        if self.gan_settings["concat_to_original_data"]:
            X_train[self.target_variable] = Y_train
            X_train = X_train.sort_index(axis=1)
            original_columns = X_train.columns
            full_new_data = np.concatenate(new_data, axis=0)
            X_train_synth = pd.DataFrame(full_new_data, columns=columns)
            X_train_synth = X_train_synth.sort_index(axis=1)
            X_train = pd.concat(
                [X_train[original_columns], X_train_synth[original_columns]]
            )
        else:
            full_new_data = np.concatenate(new_data, axis=0)
            X_train = pd.DataFrame(full_new_data, columns=columns)

        X_train = X_train.reset_index(drop=True)

        Y_train = X_train[self.target_variable]
        X_train = X_train.drop(self.target_variable, axis=1)
        print(f"Synthetic train df is of length {len(X_train.index)}")

        try:
            X_test = X_test.drop(self.target_variable, axis=1)
        except KeyError:
            pass

        del generator

        self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)



import gc
import logging
import os
import re

import numpy as np
import psutil
import torch
import torch.nn as nn
from numpy import inf
from sklearn.metrics import mean_squared_error
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from transformers import AdamW, get_linear_schedule_with_warmup

from e2eml.full_processing import cpu_preprocessing, postprocessing

# specify GPU
scaler = torch.cuda.amp.GradScaler()  # GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class RegressionDataset(Dataset):
    def __init__(self, X_data, y_data):
        self.X_data = X_data
        self.y_data = y_data

    def __getitem__(self, index):
        return self.X_data[index], self.y_data[index]

    def __len__(self):
        return len(self.X_data)


class DeepQuantileRegression(
    postprocessing.FullPipeline, cpu_preprocessing.PreProcessing, RegressionDataset
):
    """
    This class stores all model training and prediction methods for classification tasks.
    This class stores all pipeline relevant information (inherited from cpu preprocessing).
    The attribute "df_dict" always holds train and test as well as
    to predict data. The attribute "preprocess_decisions" stores encoders and other information generated during the
    model training. The attributes "predicted_classes" and "predicted_probs" store dictionaries (model names are dictionary keys)
    with predicted classes and probabilities (classification tasks) while "predicted_values" stores regression based
    predictions. The attribute "evaluation_scores" keeps track of model evaluation metrics (in dictionary format).
    :param datasource: Expects a Pandas dataframe (containing the target feature as a column)
    :param target_variable: Name of the target feature's column within the datasource dataframe.
    :param date_columns: Date columns can be passed as lists additionally for respective preprocessing. If not provided
    e2eml will try to detect datetime columns automatically. Date format is expected as YYYY-MM-DD anyway.
    :param categorical_columns: Categorical columns can be passed as lists additionally for respective preprocessing.
    If not provided e2eml will try to detect categorical columns automatically.
    :param nlp_columns: NLP columns expect a string declaring one text column.
    :param unique_identifier: A unique identifier (i.e. an ID column) can be passed as well to preserve this information
     for later processing.
    :param ml_task: Can be 'binary', 'multiclass' or 'regression'. On default will be determined automatically.
    :param preferred_training_mode: Must be 'cpu', if e2eml has been installed into an environment without LGBM and Xgboost on GPU.
    Can be set to 'gpu', if LGBM and Xgboost have been installed with GPU support. The default 'auto' will detect GPU support
    and optimize accordingly. (Default: 'auto')
    :param logging_file_path: Preferred location to save the log file. Will otherwise stored in the current folder.
    :param low_memory_mode: Adds a preprocessing feature to reduce dataframe memory footprint. Will lead to a loss in
    model performance. Will be extended by further memory savings features in future releases.
    However we highly recommend GPU usage to heavily decrease model training times.
    """

    def create_nn_train_dataset(self):
        logging.info("Create NLP train dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        train_dataset = RegressionDataset(
            torch.from_numpy(X_train.to_numpy(dtype="float32")).float(),
            torch.from_numpy(Y_train.to_numpy(dtype="float32")).float(),
        )
        return train_dataset

    def create_nn_test_dataset(self):
        logging.info("Create NLP test dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        test_dataset = RegressionDataset(
            torch.from_numpy(X_test.to_numpy(dtype="float32")).float(),
            torch.from_numpy(Y_test.to_numpy(dtype="float32")).float(),
        )
        return test_dataset

    def create_nn_pred_dataset(self):
        logging.info("Create NLP prediction dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if self.prediction_mode:
            self.dataframe[self.target_variable] = 999  # creating dummy column
            dummy_target = self.dataframe[self.target_variable]
            self.dataframe = self.dataframe.drop(self.target_variable, axis=1)
            pred_dataset = RegressionDataset(
                torch.from_numpy(self.dataframe.to_numpy()).float(),
                torch.from_numpy(dummy_target.to_numpy()).float(),
            )
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            pred_dataset = RegressionDataset(
                torch.from_numpy(X_test.to_numpy(dtype="float32")).float(),
                torch.from_numpy(Y_test.to_numpy(dtype="float32")).float(),
            )
        return pred_dataset

    def create_dual_nn_train_dataloader(self, train_batch_size=None, workers=None):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.quantile_nn_settings["train_batch_size"]

        if workers:
            pass
        else:
            workers = self.quantile_nn_settings["num_workers"]
        logging.info("Create Neural network train dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        train_dataset = self.create_nn_train_dataset()
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            num_workers=workers,
            pin_memory=True,
        )
        return train_dataloader

    def create_dual_nn_test_dataloader(self, test_batch_size=None, workers=None):
        if test_batch_size:
            pass
        else:
            test_batch_size = self.quantile_nn_settings["test_batch_size"]

        if workers:
            pass
        else:
            workers = self.quantile_nn_settings["num_workers"]
        logging.info("Create NLP test dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        test_dataset = self.create_nn_test_dataset()
        test_dataloader = DataLoader(
            test_dataset,
            batch_size=test_batch_size,
            shuffle=False,
            num_workers=workers,
            pin_memory=True,
        )
        return test_dataloader

    def pred_dual_nn_dataloader(self, pred_batch_size=None, workers=None):
        if pred_batch_size:
            pass
        else:
            pred_batch_size = self.quantile_nn_settings["pred_batch_size"]

        if workers:
            pass
        else:
            workers = self.quantile_nn_settings["num_workers"]
        logging.info("Create Neural network prediction dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        pred_dataset = self.create_nn_pred_dataset()
        pred_dataloader = DataLoader(
            pred_dataset,
            batch_size=pred_batch_size,
            shuffle=False,
            num_workers=workers,
            pin_memory=True,
        )
        return pred_dataloader

    def get_num_features(self):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        N, D = X_train.shape
        self.preprocess_decisions["num_features"] = D
        return D

    def get_dual_nn_architecture(self, type="ann", num_features=1, output_dim=1):
        if type == "ann":

            class MultipleRegression(nn.Module):
                def __init__(self, num_features, output_dim=output_dim):
                    super(MultipleRegression, self).__init__()

                    self.output_dim = output_dim
                    self.layer_1 = nn.Linear(num_features, 4096)
                    self.layer_2 = nn.Linear(4096, 512)
                    self.layer_3 = nn.Linear(512, 512)
                    self.layer_4 = nn.Linear(512, 4096)
                    self.layer_5 = nn.Linear(4096, 64)
                    self.layer_out = nn.Linear(64, self.output_dim)

                    self.relu = nn.ReLU()

                def forward(self, inputs):
                    x = self.relu(self.layer_1(inputs))
                    x = self.relu(self.layer_2(x))
                    x = self.relu(self.layer_3(x))
                    x = self.relu(self.layer_4(x))
                    x = self.relu(self.layer_5(x))
                    x = self.layer_out(x)
                    return x

                def predict(self, test_inputs):
                    x = self.relu(self.layer_1(test_inputs))
                    x = self.relu(self.layer_2(x))
                    x = self.relu(self.layer_3(x))
                    x = self.relu(self.layer_4(x))
                    x = self.relu(self.layer_5(x))
                    x = self.layer_out(x)
                    return x

            model = MultipleRegression(num_features=num_features, output_dim=3)
            return model
        else:

            class SoftOrdering1DCNN(nn.Module):
                def __init__(
                    self,
                    input_dim,
                    output_dim=output_dim,
                    sign_size=32,
                    cha_input=16,
                    cha_hidden=32,
                    K=2,
                    dropout_input=0.2,
                    dropout_hidden=0.2,
                    dropout_output=0.2,
                ):
                    super().__init__()
                    hidden_size = sign_size * cha_input
                    sign_size1 = sign_size
                    sign_size2 = sign_size // 2
                    output_size = (sign_size // 4) * cha_hidden
                    self.hidden_size = hidden_size
                    self.cha_input = cha_input
                    self.cha_hidden = cha_hidden
                    self.K = K
                    self.sign_size1 = sign_size1
                    self.sign_size2 = sign_size2
                    self.output_size = output_size
                    self.dropout_input = dropout_input
                    self.dropout_hidden = dropout_hidden
                    self.dropout_output = dropout_output
                    self.batch_norm1 = nn.BatchNorm1d(input_dim)
                    self.dropout1 = nn.Dropout(dropout_input)
                    dense1 = nn.Linear(input_dim, hidden_size, bias=False)
                    self.dense1 = nn.utils.weight_norm(dense1)
                    # 1st conv layer
                    self.batch_norm_c1 = nn.BatchNorm1d(cha_input)
                    conv1 = nn.Conv1d(
                        cha_input,
                        cha_input * K,
                        kernel_size=5,
                        stride=1,
                        padding=2,
                        groups=cha_input,
                        bias=False,
                    )
                    self.conv1 = nn.utils.weight_norm(conv1, dim=None)
                    self.ave_po_c1 = nn.AdaptiveAvgPool1d(output_size=sign_size2)
                    # 2nd conv layer
                    self.batch_norm_c2 = nn.BatchNorm1d(cha_input * K)
                    self.dropout_c2 = nn.Dropout(dropout_hidden)
                    conv2 = nn.Conv1d(
                        cha_input * K,
                        cha_hidden,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=False,
                    )
                    self.conv2 = nn.utils.weight_norm(conv2, dim=None)
                    # 3rd conv layer
                    self.batch_norm_c3 = nn.BatchNorm1d(cha_hidden)
                    self.dropout_c3 = nn.Dropout(dropout_hidden)
                    conv3 = nn.Conv1d(
                        cha_hidden,
                        cha_hidden,
                        kernel_size=3,
                        stride=1,
                        padding=1,
                        bias=False,
                    )
                    self.conv3 = nn.utils.weight_norm(conv3, dim=None)
                    # 4th conv layer
                    self.batch_norm_c4 = nn.BatchNorm1d(cha_hidden)
                    conv4 = nn.Conv1d(
                        cha_hidden,
                        cha_hidden,
                        kernel_size=5,
                        stride=1,
                        padding=2,
                        groups=cha_hidden,
                        bias=False,
                    )
                    self.conv4 = nn.utils.weight_norm(conv4, dim=None)
                    self.avg_po_c4 = nn.AvgPool1d(kernel_size=4, stride=2, padding=1)
                    self.flt = nn.Flatten()
                    self.batch_norm2 = nn.BatchNorm1d(output_size)
                    self.dropout2 = nn.Dropout(dropout_output)
                    dense2 = nn.Linear(output_size, output_dim, bias=False)
                    self.dense2 = nn.utils.weight_norm(dense2)
                    # self.loss = nn.BCEWithLogitsLoss()

                def forward(self, x):
                    x = self.batch_norm1(x)
                    x = self.dropout1(x)
                    x = nn.functional.celu(self.dense1(x))
                    x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)
                    x = self.batch_norm_c1(x)
                    x = nn.functional.relu(self.conv1(x))
                    x = self.ave_po_c1(x)
                    x = self.batch_norm_c2(x)
                    x = self.dropout_c2(x)
                    x = nn.functional.relu(self.conv2(x))
                    x_s = x
                    x = self.batch_norm_c3(x)
                    x = self.dropout_c3(x)
                    x = nn.functional.relu(self.conv3(x))
                    x = self.batch_norm_c4(x)
                    x = self.conv4(x)
                    x = x + x_s
                    x = nn.functional.relu(x)
                    x = self.avg_po_c4(x)
                    x = self.flt(x)
                    x = self.batch_norm2(x)
                    x = self.dropout2(x)
                    x = self.dense2(x)
                    return x

                def predict(self, x):
                    x = self.batch_norm1(x)
                    x = self.dropout1(x)
                    x = nn.functional.celu(self.dense1(x))
                    x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)
                    x = self.batch_norm_c1(x)
                    x = nn.functional.relu(self.conv1(x))
                    x = self.ave_po_c1(x)
                    x = self.batch_norm_c2(x)
                    x = self.dropout_c2(x)
                    x = nn.functional.relu(self.conv2(x))
                    x_s = x
                    x = self.batch_norm_c3(x)
                    x = self.dropout_c3(x)
                    x = nn.functional.relu(self.conv3(x))
                    x = self.batch_norm_c4(x)
                    x = self.conv4(x)
                    x = x + x_s
                    x = nn.functional.relu(x)
                    x = self.avg_po_c4(x)
                    x = self.flt(x)
                    x = self.batch_norm2(x)
                    x = self.dropout2(x)
                    x = self.dense2(x)
                    return x

            model = SoftOrdering1DCNN(
                input_dim=num_features,
                dropout_input=self.quantile_nn_settings["drop_out"],
                dropout_hidden=self.quantile_nn_settings["drop_out"],
                dropout_output=self.quantile_nn_settings["drop_out"],
                output_dim=3,
            )
            return model

    def dual_nn_model_setup(self, epochs=None):
        if self.prediction_mode:
            pass
        else:
            logging.info("Define NLP model.")
            logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            self.get_num_features()
            model = self.get_dual_nn_architecture(
                type=self.quantile_nn_settings["architecture"],
                num_features=self.preprocess_decisions["num_features"],
            )
            model.to(device)
            model.train()
            LR = self.quantile_nn_settings["learning_rate"]
            optimizer = torch.optim.Adam(
                model.parameters(),
                lr=LR,
                weight_decay=self.quantile_nn_settings["weight_decay"],
            )

            if epochs:
                pass
            else:
                epochs = self.quantile_nn_settings["epochs"]
            epochs = epochs
            train_steps = int(
                len(X_train) / self.quantile_nn_settings["train_batch_size"] * epochs
            )
            num_steps = int(train_steps * 0.1)
            scheduler = get_linear_schedule_with_warmup(
                optimizer, num_steps, train_steps
            )
            return model, optimizer, train_steps, num_steps, scheduler

    def simple_multi_bound_loss(self, y_hat_low, y_hat_up, target):
        mse_loss_mean = torch.mean((target - ((y_hat_low + y_hat_up) / 2)) ** 2)
        return mse_loss_mean

    def quantile_loss(self, preds, target, quantiles):
        assert not target.requires_grad
        assert preds.size(0) == target.size(0)
        losses = []
        for i, q in enumerate(quantiles):
            errors = target - preds[:, i]
            losses.append(torch.max((q - 1) * errors, q * errors).unsqueeze(1))
        loss = torch.mean(torch.sum(torch.cat(losses, dim=1), dim=1))
        return loss

    def fillna_inf(self, array):
        array = np.array(array)
        array[np.isnan(array)] = 0
        array[array == -inf] = 0
        array[array == inf] = 0
        return array

    def nn_dual_training(self, train_dataloader, model, optimizer, scheduler):
        logging.info("Start NLP training loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        model.train()
        allpreds = []
        allpreds_low = []
        allpreds_up = []
        alltargets = []
        self.reset_test_train_index(drop_target=True)

        for X_train_batch, y_train_batch in train_dataloader:
            losses = []
            optimizer.zero_grad()

            with torch.cuda.amp.autocast():
                X_train_batch, y_train_batch = X_train_batch.to(
                    device
                ), y_train_batch.to(device)

                y_train_pred = model(X_train_batch)

                y_train_batch = y_train_batch.unsqueeze(1)
                loss = self.quantile_loss(
                    y_train_pred, y_train_batch, self.quantile_nn_settings["quantiles"]
                )

                # For scoring
                losses.append(loss.item() / len(y_train_pred))
                y_train_pred = y_train_pred[:, 1:2]
                allpreds_low.append(y_train_pred[:, 0:1].detach().cpu().numpy())
                allpreds_up.append(y_train_pred[:, 2:3].detach().cpu().numpy())
                allpreds.append(y_train_pred.detach().cpu().numpy())
                alltargets.append(y_train_batch.detach().squeeze(-1).cpu().numpy())

            scaler.scale(loss).backward()  # backwards of loss
            scaler.step(optimizer)  # Update optimizer
            scaler.update()  # scaler update
            scheduler.step()  # Update learning rate schedule

            # Combine dataloader minutes

        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        allpreds = self.fillna_inf(allpreds)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        try:
            train_rme_loss = mean_squared_error(alltargets, allpreds)
        except ValueError:
            train_rme_loss = mean_squared_error(alltargets, allpreds)
        return losses, train_rme_loss

    def nn_dual_validating(self, valid_dataloader, model):
        logging.info("Start NLP validation loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        model.eval()
        allpreds = []
        allpreds_low = []
        allpreds_up = []
        alltargets = []

        for X_test_batch, y_test_batch in valid_dataloader:
            losses = []

            with torch.no_grad():
                X_test_batch, y_test_batch = X_test_batch.to(device), y_test_batch.to(
                    device
                )
                y_train_pred = model(X_test_batch)
                y_test_batch = y_test_batch.unsqueeze(1)
                loss = self.quantile_loss(
                    y_train_pred, y_test_batch, self.quantile_nn_settings["quantiles"]
                )

                # For scoring
                losses.append(loss.item() / len(y_train_pred))
                y_train_pred = y_train_pred[:, 1:2]
                allpreds_low.append(y_train_pred[:, 0:1].detach().cpu().numpy())
                allpreds_up.append(y_train_pred[:, 2:3].detach().cpu().numpy())
                allpreds.append(y_train_pred.detach().cpu().numpy())
                alltargets.append(y_test_batch.detach().squeeze(-1).cpu().numpy())

        allpreds_low = np.concatenate(allpreds_low)
        allpreds_up = np.concatenate(allpreds_up)
        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        allpreds_low = self.fillna_inf(allpreds_low)
        allpreds_up = self.fillna_inf(allpreds_up)
        allpreds = self.fillna_inf(allpreds)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        try:
            valid_rme_loss = mean_squared_error(alltargets, allpreds)
            # valid_rme_loss = np.nan_to_num(valid_rme_loss, nan=0, neginf=0)
        except ValueError:
            allpreds_low = self.fillna_inf(allpreds_low)
            allpreds_up = self.fillna_inf(allpreds_up)
            valid_rme_loss = mean_squared_error(alltargets, allpreds)

        return allpreds, losses, valid_rme_loss

    def nn_dual_predicting(self, pred_dataloader, model, pathes):
        logging.info("Start Neural network prediction loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        allpreds = []
        model_no = 0
        mode_cols = []
        for m_path in pathes:
            state = torch.load(m_path)
            model.load_state_dict(state["state_dict"])
            model.to(device)
            model.eval()
            preds = []
            preds_low = []
            preds_up = []
            # allvalloss = 0
            with torch.no_grad():
                for X_pred_batch, y_pred_batch in pred_dataloader:
                    X_pred_batch, y_pred_batch = X_pred_batch.to(
                        device
                    ), y_pred_batch.to(device)
                    y_pred_pred = model(X_pred_batch)

                    preds_low.append(y_pred_pred[:, 0:1].detach().cpu().numpy())
                    preds_up.append(y_pred_pred[:, 2:3].detach().cpu().numpy())
                    preds.append(y_pred_pred[:, 1:2].detach().cpu().numpy())

                preds = np.concatenate(preds)
                preds_low = np.concatenate(preds_low)
                preds_up = np.concatenate(preds_up)

                if self.prediction_mode:
                    self.dataframe[f"preds_model{model_no}"] = preds
                    self.dataframe[f"preds_low_model{model_no}"] = preds_low
                    self.dataframe[f"preds_up_model{model_no}"] = preds_up

                    self.dataframe[
                        f"preds_model{model_no}"
                    ] = self.data_scaling_target_only(
                        mode="reverse",
                        to_scale=self.dataframe[f"preds_model{model_no}"],
                    )
                    self.dataframe[
                        f"preds_low_model{model_no}"
                    ] = self.data_scaling_target_only(
                        mode="reverse",
                        to_scale=self.dataframe[f"preds_low_model{model_no}"],
                    )
                    self.dataframe[
                        f"preds_up_model{model_no}"
                    ] = self.data_scaling_target_only(
                        mode="reverse",
                        to_scale=self.dataframe[f"preds_up_model{model_no}"],
                    )

                    self.dataframe[
                        f"preds_low_model{model_no}"
                    ] = self.target_skewness_handling(
                        preds_to_reconvert=self.dataframe[
                            f"preds_low_model{model_no}"
                        ].values,
                        mode="revert",
                    )
                    self.dataframe[
                        f"preds_up_model{model_no}"
                    ] = self.target_skewness_handling(
                        preds_to_reconvert=self.dataframe[
                            f"preds_up_model{model_no}"
                        ].values,
                        mode="revert",
                    )
                else:
                    X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
                    X_test[f"preds_model{model_no}"] = preds
                    X_test[f"preds_low_model{model_no}"] = preds_low
                    X_test[f"preds_up_model{model_no}"] = preds_up

                    X_test[f"preds_model{model_no}"] = self.data_scaling_target_only(
                        mode="reverse", to_scale=X_test[f"preds_model{model_no}"]
                    )
                    X_test[
                        f"preds_low_model{model_no}"
                    ] = self.data_scaling_target_only(
                        mode="reverse", to_scale=X_test[f"preds_low_model{model_no}"]
                    )
                    X_test[f"preds_up_model{model_no}"] = self.data_scaling_target_only(
                        mode="reverse", to_scale=X_test[f"preds_up_model{model_no}"]
                    )

                    X_test[
                        f"preds_low_model{model_no}"
                    ] = self.target_skewness_handling(
                        preds_to_reconvert=X_test[f"preds_low_model{model_no}"].values,
                        mode="revert",
                    )
                    X_test[f"preds_up_model{model_no}"] = self.target_skewness_handling(
                        preds_to_reconvert=X_test[f"preds_up_model{model_no}"].values,
                        mode="revert",
                    )
                    self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)
                mode_cols.append(f"preds_model{model_no}")

                allpreds.append(preds)
                model_no += 1
            del state
            torch.cuda.empty_cache()
            _ = gc.collect()
        return allpreds, mode_cols

    def load_model_states(self, path=None):
        logging.info("Load model save states.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if path:
            pass
        else:
            path = os.getcwd()
        if self.prediction_mode:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if ".pth" in s
            ]
            return pthes
        else:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if ".pth" in s
            ]
            return pthes

    def deep_quantile_regression_train(self):
        if self.prediction_mode:
            pass
        else:
            logging.info("Start NLP transformer training.")
            logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
            self.reset_test_train_index(drop_target=True)

            train_dataloader = self.create_dual_nn_train_dataloader()
            test_dataloader = self.create_dual_nn_test_dataloader()
            (
                model,
                optimizer,
                train_steps,
                num_steps,
                scheduler,
            ) = self.dual_nn_model_setup()
            scheduler = get_linear_schedule_with_warmup(
                optimizer, num_steps, train_steps
            )

            trainlosses = []
            vallosses = []
            bestscore = None
            trainscores = []
            validscores = []

            for epoch in tqdm(range(self.quantile_nn_settings["epochs"])):
                print("---------------" + str(epoch) + "start-------------")
                trainloss, trainscore = self.nn_dual_training(
                    train_dataloader, model, optimizer, scheduler
                )
                trainlosses.append(trainloss)
                trainscores.append(trainscore)
                print("trainscore is " + str(trainscore))
                preds, validloss, valscore = self.nn_dual_validating(
                    test_dataloader, model
                )
                vallosses.append(validloss)
                validscores.append(valscore)

                print("valscore is " + str(valscore))
                if bestscore is None:
                    bestscore = valscore
                    print("Save first model")
                    state = {
                        "state_dict": model.state_dict(),
                        "optimizer_dict": optimizer.state_dict(),
                        "bestscore": bestscore,
                    }
                    torch.save(state, "model0.pth")

                elif bestscore > valscore:
                    bestscore = valscore
                    print("found better point")
                    state = {
                        "state_dict": model.state_dict(),
                        "optimizer_dict": optimizer.state_dict(),
                        "bestscore": bestscore,
                    }
                    torch.save(state, "model0.pth")
                else:
                    pass

            bestscores = []
            bestscores.append(bestscore)

            for fold in range(self.quantile_nn_settings["nb_model_to_create"]):

                self.reset_test_train_index(drop_target=True)
                X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

                # initializing the data
                train_dataloader = self.create_dual_nn_train_dataloader()
                test_dataloader = self.create_dual_nn_test_dataloader()

                model = self.get_dual_nn_architecture(
                    type=self.quantile_nn_settings["architecture"],
                    num_features=self.preprocess_decisions["num_features"],
                )
                model.to(device)
                LR = 2e-5  # 1e-3
                optimizer = AdamW(
                    model.parameters(), LR, betas=(0.99, 0.999), weight_decay=1e-2
                )
                train_steps = int(
                    len(X_train)
                    / self.quantile_nn_settings["train_batch_size"]
                    * self.quantile_nn_settings["epochs"]
                )
                num_steps = int(train_steps * 0.1)
                scheduler = get_linear_schedule_with_warmup(
                    optimizer, num_steps, train_steps
                )

                trainlosses = []
                vallosses = []
                bestscore = None
                trainscores = []
                validscores = []

                for epoch in tqdm(range(self.quantile_nn_settings["epochs"])):
                    print("---------------" + str(epoch) + "start-------------")
                    trainloss, trainscore = self.nn_dual_training(
                        train_dataloader, model, optimizer, scheduler
                    )
                    trainlosses.append(trainloss)
                    trainscores.append(trainscore)

                    print("trainscore is " + str(trainscore))
                    preds, validloss, valscore = self.nn_dual_validating(
                        test_dataloader, model
                    )

                    vallosses.append(validloss)
                    validscores.append(valscore)
                    print("valscore is " + str(valscore))

                    if bestscore is None:
                        bestscore = valscore
                        print("Save first model")
                        state = {
                            "state_dict": model.state_dict(),
                            "optimizer_dict": optimizer.state_dict(),
                            "bestscore": bestscore,
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    elif bestscore > valscore:
                        bestscore = valscore
                        print("found better point")
                        state = {
                            "state_dict": model.state_dict(),
                            "optimizer_dict": optimizer.state_dict(),
                            "bestscore": bestscore,
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    else:
                        pass
                bestscores.append(bestscore)

            del model, optimizer, scheduler
            _ = gc.collect()

    def deep_quantile_regression_predict(self):
        logging.info("Start NLP transformer prediction.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        self.reset_test_train_index(drop_target=True)
        model = self.get_dual_nn_architecture(
            type=self.quantile_nn_settings["architecture"],
            num_features=self.preprocess_decisions["num_features"],
        )
        pathes = self.load_model_states(path=self.tabular_nn_model_save_states_path)
        pthes = self.load_model_states(path=self.tabular_nn_model_save_states_path)
        try:
            for path in pathes:
                if re.search("generator_model.pth", path):
                    pthes.remove(path)
                if re.search("autoencoder_model.pth", path):
                    pthes.remove(path)
        except Exception:
            pass
        print(pthes)
        pred_dataloader = self.pred_dual_nn_dataloader()
        allpreds, mode_cols = self.nn_dual_predicting(pred_dataloader, model, pthes)

        if self.prediction_mode:
            self.dataframe["deep_quantile_regression_median"] = self.dataframe[
                mode_cols
            ].median(axis=1)
            self.dataframe["deep_quantile_regression_mean"] = self.dataframe[
                mode_cols
            ].mean(axis=1)
            self.dataframe[
                "deep_quantile_regression_mean"
            ] = self.target_skewness_handling(
                preds_to_reconvert=self.dataframe[
                    "deep_quantile_regression_mean"
                ].values,
                mode="revert",
            )
            self.predicted_values["deep_quantile_regression"] = self.dataframe[
                "deep_quantile_regression_mean"
            ]

        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

            X_test["deep_quantile_regression_median"] = X_test[mode_cols].median(axis=1)
            X_test["deep_quantile_regression_mean"] = X_test[mode_cols].mean(axis=1)
            X_test["deep_quantile_regression_mean"] = self.target_skewness_handling(
                preds_to_reconvert=X_test["deep_quantile_regression_mean"].values,
                mode="revert",
            )
            self.predicted_values["deep_quantile_regression"] = X_test[
                "deep_quantile_regression_median"
            ]
            self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)



import gc
import logging
import warnings

import numpy as np
import optuna
import pandas as pd
import torch
from pandas.core.common import SettingWithCopyWarning
from pytorch_tabnet.pretraining import TabNetPretrainer
from pytorch_tabnet.tab_model import TabNetRegressor
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import TimeSeriesSplit

try:
    from statsmodels.tsa.arima.model import ARIMA
    from statsmodels.tsa.holtwinters import ExponentialSmoothing
except ModuleNotFoundError:
    from statsmodels.tsa.arima_model import ARIMA

from torch.optim.lr_scheduler import ReduceLROnPlateau

from e2eml.full_processing import postprocessing

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class UnivariateTimeSeriesModels(postprocessing.FullPipeline):
    """
    This class stores all model training and prediction methods for regression tasks.
    This class stores all pipeline relevant information (inherited from cpu preprocessing).
    The attribute "df_dict" always holds train and test as well as
    to predict data. The attribute "preprocess_decisions" stores encoders and other information generated during the
    model training. The attributes "predicted_classes" and "predicted_probs" store dictionaries (model names are dictionary keys)
    with predicted classes and probabilities (classification tasks) while "predicted_values" stores regression based
    predictions. The attribute "evaluation_scores" keeps track of model evaluation metrics (in dictionary format).
    :param datasource: Expects a Pandas dataframe (containing the target feature as a column)
    :param target_variable: Name of the target feature's column within the datasource dataframe.
    :param date_columns: Date columns can be passed as lists additionally for respective preprocessing. If not provided
    e2eml will try to detect datetime columns automatically. Date format is expected as YYYY-MM-DD anyway.
    :param categorical_columns: Categorical columns can be passed as lists additionally for respective preprocessing.
    If not provided e2eml will try to detect categorical columns automatically.
    :param nlp_columns: NLP columns can be passed specifically. This only makes sense, if the chosen blueprint runs under 'nlp' processing.
    If NLP columns are not declared, categorical columns will be interpreted as such.
    :param unique_identifier: A unique identifier (i.e. an ID column) can be passed as well to preserve this information
     for later processing.
    :param ml_task: Can be 'binary', 'multiclass' or 'regression'. On default will be determined automatically.
    :param preferred_training_mode: Must be 'cpu', if e2eml has been installed into an environment without LGBM and Xgboost on GPU.
    Can be set to 'gpu', if LGBM and Xgboost have been installed with GPU support. The default 'auto' will detect GPU support
    and optimize accordingly. Only TabNet can only run on GPU and will not be impacted from this parameter. (Default: 'auto')
    :param logging_file_path: Preferred location to save the log file. Will otherwise stored in the current folder.
    :param low_memory_mode: Adds a preprocessing feature to reduce dataframe memory footprint. Will lead to a loss in
    model performance. Will be extended by further memory savings features in future releases.
    However we highly recommend GPU usage to heavily decrease model training times.
    """

    def auto_arima_train(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            algorithm = "auto_arima"
            tscv = TimeSeriesSplit(n_splits=5)

            def objective(trial):
                if self.preprocess_decisions["arima_transformation_used"] == "None":
                    i_term = self.preprocess_decisions["arima_nb_differentiations"]
                else:
                    i_term = 0

                param = {
                    "p": trial.suggest_int("p", 1, self.auto_arima_settings["max_p"]),
                    "i": i_term,
                    "q": trial.suggest_int("q", 1, self.auto_arima_settings["max_q"]),
                }

                mean_abs_errors = []
                for train_index, test_index in tscv.split(X_train):
                    x_train, x_test = (
                        X_train.iloc[train_index],
                        X_train.iloc[test_index],
                    )
                    y_train, y_test = (  # noqa: F841
                        Y_train.iloc[train_index],
                        Y_train.iloc[test_index],
                    )
                    model = ARIMA(
                        x_train.values, order=(param["p"], param["i"], param["q"])
                    )
                    try:
                        model = model.fit()
                        preds = model.forecast(len(x_test.index))
                        if (
                            self.preprocess_decisions["arima_transformation_used"]
                            == "log"
                        ):
                            y_test = np.expm1(y_test.astype(float))
                            preds = np.expm1(preds.astype(float))
                        mae = mean_absolute_error(y_test, preds)
                        mean_abs_errors.append(mae)
                    except Exception as e:
                        mae = 9999999999
                        mean_abs_errors.append(mae)
                        print(e)
                return np.mean(np.asarray(mean_abs_errors))

            sampler = optuna.samplers.TPESampler(
                multivariate=True, seed=self.global_random_state
            )
            study = optuna.create_study(
                direction="minimize", sampler=sampler, study_name=f"{algorithm}"
            )
            study.optimize(
                objective,
                n_trials=self.hyperparameter_tuning_rounds[algorithm],
                timeout=self.hyperparameter_tuning_max_runtime_secs[algorithm],
                gc_after_trial=True,
                show_progress_bar=True,
            )
            self.optuna_studies[f"{algorithm}"] = {}
            # optuna.visualization.plot_optimization_history(study).write_image('LGBM_optimization_history.png')
            # optuna.visualization.plot_param_importances(study).write_image('LGBM_param_importances.png')
            try:
                fig = optuna.visualization.plot_optimization_history(study)
                self.optuna_studies[f"{algorithm}_plot_optimization"] = fig
                fig.show()
                fig = optuna.visualization.plot_param_importances(study)
                self.optuna_studies[f"{algorithm}_param_importance"] = fig
                fig.show()
            except ZeroDivisionError:
                pass

            best_parameters = study.best_trial.params
            if self.preprocess_decisions["arima_transformation_used"] == "None":
                best_parameters["i"] = 0
            elif self.preprocess_decisions["arima_transformation_used"] == "log":
                best_parameters["i"] = 0
            elif self.preprocess_decisions["arima_nb_differentiations"] > 0:
                best_parameters["i"] = self.preprocess_decisions[
                    "arima_nb_differentiations"
                ]

            model = ARIMA(
                pd.concat([X_train, X_test]).values,
                order=(
                    best_parameters["p"],
                    best_parameters["i"],
                    best_parameters["q"],
                ),
            )
            model = model.fit()
            self.trained_models[f"{algorithm}"] = {}
            self.trained_models[f"{algorithm}"] = model
            del model
            _ = gc.collect()
            return self.trained_models

    def auto_arima_predict(self, n_forecast=1):
        """
        Loads the pretrained model from the class itself and predicts on new data.
        :return: Updates class attributes.
        """
        self.get_current_timestamp(task="Predict with Auto Arima")
        algorithm = "auto_arima"
        if self.prediction_mode:
            model = self.trained_models[f"{algorithm}"]
            predicted_values = model.forecast(n_forecast)
            predicted_values[np.isfinite(predicted_values) is False] = 0
            if self.preprocess_decisions["arima_transformation_used"] == "log":
                predicted_values = np.expm1(predicted_values.astype(float))
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            model = self.trained_models[f"{algorithm}"]
            predicted_values = model.forecast(len(X_test.index))
            predicted_values[np.isfinite(predicted_values) is False] = 0
            if self.preprocess_decisions["arima_transformation_used"] == "log":
                predicted_values = np.expm1(predicted_values.astype(float))
        self.predicted_values[f"{algorithm}"] = {}
        self.predicted_values[f"{algorithm}"] = predicted_values
        del model
        _ = gc.collect()

    def holt_winters_train(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            algorithm = "holt_winters"
            tscv = TimeSeriesSplit(n_splits=3)
            stric_pos = (X_train == 0).sum().sum() == 0

            if not stric_pos:
                delta = 0 - X_train.min()
                X_train = X_train + delta + 0.01
                self.preprocess_decisions["holt_winters_delta"] = delta.values[0]

            stric_pos = (X_train == 0).sum().sum() == 0
            print((X_train == 0).sum().sum())

            def objective(trial):
                options = ["additive", "multiplicative", None]
                damped_trend_options = [True, False]
                if not stric_pos:
                    options.remove("additive")
                    options.remove("multiplicative")
                    damped_trend_options.remove(True)
                param = {
                    "trend": trial.suggest_categorical("trend", options),
                    "damped_trend": trial.suggest_categorical(
                        "damped_trend", damped_trend_options
                    ),
                    "seasonal": trial.suggest_categorical("seasonal", options),
                    "initialization_method": trial.suggest_categorical(
                        "initialization_method", ["estimated", None]
                    ),
                    "use_boxcox": trial.suggest_categorical(
                        "use_boxcox", [True, False]
                    ),
                    "method": trial.suggest_categorical(
                        "method",
                        [
                            "L-BFGS-B",
                            "TNC",
                            "SLSQP",
                            "Powell",
                            "trust-constr",
                            "basinhopping",
                        ],
                    ),
                }

                if param["trend"]:
                    param["damped_trend"] = trial.suggest_categorical(
                        "damped_trend", [True, False]
                    )
                else:
                    param["damped_trend"] = False

                mean_abs_errors = []
                for train_index, test_index in tscv.split(X_train):
                    x_train, x_test = (
                        X_train.iloc[train_index],
                        X_train.iloc[test_index],
                    )

                    y_train, y_test = (  # noqa: F841
                        Y_train.iloc[train_index],
                        Y_train.iloc[test_index],
                    )
                    max_len = int((len(x_train.index)) / 3)
                    if param["seasonal"] in [
                        "add",
                        "mul",
                        "additive",
                        "multiplicative",
                    ]:
                        param["seasonal_periods"] = trial.suggest_int(
                            "seasonal_periods", 2, max_len
                        )
                    else:
                        param["seasonal_periods"] = None

                    model = ExponentialSmoothing(
                        x_train,
                        trend=param["trend"],
                        damped_trend=param["damped_trend"],
                        seasonal=param["seasonal"],
                        initialization_method=param["initialization_method"],
                        seasonal_periods=param["seasonal_periods"],
                    )
                    model = model.fit(
                        optimized=True, method=param["method"], use_brute=True
                    )
                    preds = model.forecast(len(x_test.index))
                    mae = mean_absolute_error(
                        y_test, preds + self.preprocess_decisions["holt_winters_delta"]
                    )
                    mean_abs_errors.append(mae)

                return np.mean(np.asarray(mean_abs_errors))

            sampler = optuna.samplers.TPESampler(
                multivariate=True, seed=self.global_random_state
            )
            study = optuna.create_study(
                direction="minimize", sampler=sampler, study_name=f"{algorithm}"
            )
            study.optimize(
                objective,
                n_trials=self.hyperparameter_tuning_rounds[algorithm],
                timeout=self.hyperparameter_tuning_max_runtime_secs[algorithm],
                gc_after_trial=True,
                show_progress_bar=True,
            )
            self.optuna_studies[f"{algorithm}"] = {}
            # optuna.visualization.plot_optimization_history(study).write_image('LGBM_optimization_history.png')
            # optuna.visualization.plot_param_importances(study).write_image('LGBM_param_importances.png')
            try:
                fig = optuna.visualization.plot_optimization_history(study)
                self.optuna_studies[f"{algorithm}_plot_optimization"] = fig
                fig.show()
                fig = optuna.visualization.plot_param_importances(study)
                self.optuna_studies[f"{algorithm}_param_importance"] = fig
                fig.show()
            except ZeroDivisionError:
                pass

            best_parameters = study.best_trial.params

            if "seasonal_periods" not in best_parameters:
                best_parameters["seasonal_periods"] = None

            model = ExponentialSmoothing(
                pd.concat([X_train, X_test]).values,
                trend=best_parameters["trend"],
                damped_trend=best_parameters["damped_trend"],
                seasonal=best_parameters["seasonal"],
                initialization_method=best_parameters["initialization_method"],
                seasonal_periods=best_parameters["seasonal_periods"],
            )
            model = model.fit(
                optimized=True, method=best_parameters["method"], use_brute=True
            )
            self.trained_models[f"{algorithm}"] = {}
            self.trained_models[f"{algorithm}"] = model
            del model
            _ = gc.collect()
            return self.trained_models

    def holt_winters_predict(self, n_forecast=1):
        """
        Loads the pretrained model from the class itself and predicts on new data.
        :param feat_importance: Set True, if feature importance shall be calculated.
        :param importance_alg: Chose 'permutation' (recommended on CPU) or 'SHAP' (recommended when model uses
        GPU acceleration). (Default: 'permutation')
        :return: Updates class attributes.
        """
        self.get_current_timestamp(task="Predict with Holt winters")
        algorithm = "holt_winters"
        if self.prediction_mode:
            self.reset_test_train_index()
            model = self.trained_models[f"{algorithm}"]
            predicted_probs = model.forecast(n_forecast)

        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            model = self.trained_models[f"{algorithm}"]
            predicted_probs = model.forecast(len(X_test.index))

        self.predicted_values[f"{algorithm}"] = {}
        self.predicted_values[f"{algorithm}"] = (
            predicted_probs + self.preprocess_decisions["holt_winters_delta"]
        )
        del model
        _ = gc.collect()


class RegressionForTimeSeriesModels(postprocessing.FullPipeline):
    """
    This class stores all model training and prediction methods for regression tasks.
    This class stores all pipeline relevant information (inherited from cpu preprocessing).
    The attribute "df_dict" always holds train and test as well as
    to predict data. The attribute "preprocess_decisions" stores encoders and other information generated during the
    model training. The attributes "predicted_classes" and "predicted_probs" store dictionaries (model names are dictionary keys)
    with predicted classes and probabilities (classification tasks) while "predicted_values" stores regression based
    predictions. The attribute "evaluation_scores" keeps track of model evaluation metrics (in dictionary format).
    :param datasource: Expects a Pandas dataframe (containing the target feature as a column)
    :param target_variable: Name of the target feature's column within the datasource dataframe.
    :param date_columns: Date columns can be passed as lists additionally for respective preprocessing. If not provided
    e2eml will try to detect datetime columns automatically. Date format is expected as YYYY-MM-DD anyway.
    :param categorical_columns: Categorical columns can be passed as lists additionally for respective preprocessing.
    If not provided e2eml will try to detect categorical columns automatically.
    :param nlp_columns: NLP columns can be passed specifically. This only makes sense, if the chosen blueprint runs under 'nlp' processing.
    If NLP columns are not declared, categorical columns will be interpreted as such.
    :param unique_identifier: A unique identifier (i.e. an ID column) can be passed as well to preserve this information
     for later processing.
    :param ml_task: Can be 'binary', 'multiclass' or 'regression'. On default will be determined automatically.
    :param preferred_training_mode: Must be 'cpu', if e2eml has been installed into an environment without LGBM and Xgboost on GPU.
    Can be set to 'gpu', if LGBM and Xgboost have been installed with GPU support. The default 'auto' will detect GPU support
    and optimize accordingly. Only TabNet can only run on GPU and will not be impacted from this parameter. (Default: 'auto')
    :param logging_file_path: Preferred location to save the log file. Will otherwise stored in the current folder.
    :param low_memory_mode: Adds a preprocessing feature to reduce dataframe memory footprint. Will lead to a loss in
    model performance. Will be extended by further memory savings features in future releases.
    However we highly recommend GPU usage to heavily decrease model training times.
    """

    def tabnet_regression_train(self):
        """
        Trains a simple Linear regression classifier.
        :return: Trained model.
        """
        self.get_current_timestamp(task="Train Tabnet regression model")
        algorithm = "tabnet"
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            x_train_sample, y_train_sample = self.get_hyperparameter_tuning_sample_df()

            X_train[X_train.columns.to_list()] = X_train[
                X_train.columns.to_list()
            ].astype(float)
            X_test[X_test.columns.to_list()] = X_test[X_test.columns.to_list()].astype(
                float
            )
            x_train_sample[x_train_sample.columns.to_list()] = x_train_sample[
                x_train_sample.columns.to_list()
            ].astype(float)

            y_train_sample = y_train_sample.astype(int)

            if len(x_train_sample.index) < len(X_train.index):
                rec_batch_size = (len(x_train_sample.index) * 0.8) / 20
                if int(rec_batch_size) % 2 == 0:
                    rec_batch_size = int(rec_batch_size)
                else:
                    rec_batch_size = int(rec_batch_size) + 1
                if rec_batch_size > 16384:
                    rec_batch_size = 16384
                    virtual_batch_size = 4096
                else:
                    virtual_batch_size = int(rec_batch_size / 4)

                # update batch sizes in case hyperparameter tuning happens on samples
                self.tabnet_settings["batch_size"] = rec_batch_size
                self.tabnet_settings["virtual_batch_size"] = virtual_batch_size

            # load settings
            batch_size = self.tabnet_settings["batch_size"]
            virtual_batch_size = self.tabnet_settings["virtual_batch_size"]
            num_workers = self.tabnet_settings["num_workers"]

            def objective(trial):
                depths = trial.suggest_int("depths", 16, 64)
                factor = trial.suggest_uniform("factor", 0.1, 0.9)
                pretrain_difficulty = trial.suggest_uniform(
                    "pretrain_difficulty", 0.7, 0.9
                )
                mode = trial.suggest_categorical("mode", ["max", "min"])
                gamma = trial.suggest_loguniform("gamma", 1e-5, 2.0)
                lambda_sparse = trial.suggest_loguniform("lambda_sparse", 1e-6, 1e-3)
                mask_type = trial.suggest_categorical(
                    "mask_type", ["sparsemax", "entmax"]
                )
                n_shared = trial.suggest_int("n_shared", 1, 5)
                n_independent = trial.suggest_int("n_independent", 1, 5)
                pre_trainer_max_epochs = trial.suggest_int(
                    "pre_trainer_max_epochs", 10, 1000
                )
                trainer_max_epochs = trial.suggest_int("trainer_max_epochs", 10, 1000)
                # loss_func = trial.suggest_categorical('loss_func', ['rmsle', 'mae', 'rmse', 'mse'])
                # ['auc', 'accuracy', 'balanced_accuracy', 'logloss', 'mae', 'mse', 'rmsle', 'unsup_loss', 'rmse']"

                param = dict(
                    gamma=gamma,
                    lambda_sparse=lambda_sparse,
                    n_d=depths,
                    n_a=depths,
                    n_shared=n_shared,
                    n_independent=n_independent,
                    n_steps=trial.suggest_int("n_steps", 1, 5),
                    optimizer_params=dict(lr=2e-2, weight_decay=1e-5),
                    mask_type=mask_type,
                    scheduler_params=dict(
                        mode=mode, patience=50, min_lr=1e-5, factor=factor
                    ),
                    scheduler_fn=ReduceLROnPlateau,
                    seed=self.global_random_state,
                    verbose=1,
                )
                mean_abs_errors = []
                skf = TimeSeriesSplit(n_splits=10)

                for train_index, test_index in skf.split(x_train_sample):
                    x_train, x_test = (
                        x_train_sample.iloc[train_index],
                        x_train_sample.iloc[test_index],
                    )
                    y_train, y_test = (
                        y_train_sample.iloc[train_index],
                        y_train_sample.iloc[test_index],
                    )
                    # numpy conversion
                    y_train = y_train.values.reshape(-1, 1)
                    y_test = y_test.values.reshape(-1, 1)
                    x_train = x_train.to_numpy()
                    x_test = x_test.to_numpy()

                    Y_train_num = y_train_sample.values.reshape(-1, 1)  # noqa: F841
                    Y_test_num = Y_test.values.reshape(-1, 1)
                    X_train_num = x_train_sample.to_numpy()  # noqa: F841
                    X_test_num = X_test.to_numpy()

                    pretrainer = TabNetPretrainer(**param)
                    pretrainer.fit(
                        x_train,
                        eval_set=[(x_test)],
                        max_epochs=pre_trainer_max_epochs,
                        patience=50,
                        batch_size=batch_size,
                        virtual_batch_size=virtual_batch_size,
                        num_workers=num_workers,
                        drop_last=True,
                        pretraining_ratio=pretrain_difficulty,
                    )

                    model = TabNetRegressor(**param)
                    model.fit(
                        x_train,
                        y_train,
                        eval_set=[(x_test, y_test)],
                        eval_metric=["mae"],
                        patience=50,
                        batch_size=batch_size,
                        virtual_batch_size=virtual_batch_size,
                        num_workers=num_workers,
                        max_epochs=trainer_max_epochs,
                        drop_last=True,
                        from_unsupervised=pretrainer,
                    )
                    preds = model.predict(X_test_num)

                    mae = mean_absolute_error(Y_test_num, preds)
                    mean_abs_errors.append(mae)
                cv_mae = np.mean(mean_abs_errors)
                return cv_mae

            study = optuna.create_study(
                direction="minimize", study_name=f"{algorithm} tuning"
            )

            logging.info("Start Tabnet validation.")

            study.optimize(
                objective,
                n_trials=self.hyperparameter_tuning_rounds["tabnet"],
                timeout=self.hyperparameter_tuning_max_runtime_secs["tabnet"],
                gc_after_trial=True,
                show_progress_bar=True,
            )
            self.optuna_studies[f"{algorithm}"] = {}
            # optuna.visualization.plot_optimization_history(study).write_image('LGBM_optimization_history.png')
            # optuna.visualization.plot_param_importances(study).write_image('LGBM_param_importances.png')
            try:
                fig = optuna.visualization.plot_optimization_history(study)
                self.optuna_studies[f"{algorithm}_plot_optimization"] = fig
                fig.show()
                fig = optuna.visualization.plot_param_importances(study)
                self.optuna_studies[f"{algorithm}_param_importance"] = fig
                fig.show()
            except ZeroDivisionError:
                pass

            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            Y_train = Y_train.values.reshape(-1, 1)
            Y_test = Y_test.values.reshape(-1, 1)
            X_train = X_train.to_numpy()
            X_test = X_test.to_numpy()

            tabnet_best_param = study.best_trial.params
            param = dict(
                gamma=tabnet_best_param["gamma"],
                lambda_sparse=tabnet_best_param["lambda_sparse"],
                n_d=tabnet_best_param["depths"],
                n_a=tabnet_best_param["depths"],
                n_shared=tabnet_best_param["n_shared"],
                n_independent=tabnet_best_param["n_independent"],
                n_steps=tabnet_best_param["n_steps"],
                optimizer_params=dict(lr=2e-2, weight_decay=1e-5),
                mask_type=tabnet_best_param["mask_type"],
                scheduler_params=dict(
                    mode=tabnet_best_param["mode"],
                    patience=5,
                    min_lr=1e-5,
                    factor=tabnet_best_param["factor"],
                ),
                scheduler_fn=ReduceLROnPlateau,
                seed=self.global_random_state,
                verbose=1,
            )
            pretrainer = TabNetPretrainer(**param)
            pretrainer.fit(
                X_train,
                eval_set=[(X_test)],
                max_epochs=tabnet_best_param["pre_trainer_max_epochs"],
                patience=50,
                batch_size=batch_size,
                virtual_batch_size=virtual_batch_size,
                num_workers=num_workers,
                drop_last=True,
                pretraining_ratio=tabnet_best_param["pretrain_difficulty"],
            )

            model = TabNetRegressor(**param)

            model.fit(
                X_train,
                Y_train,
                eval_set=[(X_test, Y_test)],
                eval_metric=["mae"],
                patience=50,
                batch_size=batch_size,
                virtual_batch_size=virtual_batch_size,
                num_workers=num_workers,
                max_epochs=tabnet_best_param["trainer_max_epochs"],
                drop_last=True,
                from_unsupervised=pretrainer,
            )
            self.trained_models[f"{algorithm}"] = {}
            self.trained_models[f"{algorithm}"] = model
            del model
            _ = gc.collect()
            return self.trained_models

    def tabnet_regression_predict(self):
        """
        Loads the pretrained model from the class itself and predicts on new data.
        :param feat_importance: Set True, if feature importance shall be calculated.
        :param importance_alg: Chose 'permutation' (recommended on CPU) or 'SHAP' (recommended when model uses
        GPU acceleration). (Default: 'permutation')
        :return: Updates class attributes.
        """
        self.get_current_timestamp(task="Predict with Tabnet regression")
        algorithm = "tabnet"
        if self.prediction_mode:
            self.reset_test_train_index()
            model = self.trained_models[f"{algorithm}"]
            predicted_probs = model.predict(self.dataframe.to_numpy(dtype="float32"))
            # self.dataframe[self.target_variable] = predicted_probs
            # self.scale_with_target(mode="reverse", drop_target=True)
            predicted_probs = self.target_skewness_handling(
                preds_to_reconvert=predicted_probs, mode="revert"
            )
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            model = self.trained_models[f"{algorithm}"]
            predicted_probs = model.predict(X_test.to_numpy(dtype="float32"))
            # X_test[self.target_variable] = predicted_probs
            # self.scale_with_target(mode="reverse", drop_target=True)
            predicted_probs = self.target_skewness_handling(
                preds_to_reconvert=predicted_probs, mode="revert"
            )

        self.predicted_values[f"{algorithm}"] = {}
        self.predicted_values[f"{algorithm}"] = predicted_probs
        del model
        _ = gc.collect()



import gc
import logging
import os
import re

import numpy as np
import psutil
import torch
import torch.nn as nn
from numpy import inf
from sklearn.metrics import mean_squared_error
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from transformers import AdamW, get_linear_schedule_with_warmup

from e2eml.full_processing import postprocessing

# specify GPU
scaler = torch.cuda.amp.GradScaler()  # GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class TimeSeriesDataset(Dataset):
    def __init__(self, dataframe, target, features, sequence_length=5):
        self.features = features
        self.target = target
        self.sequence_length = sequence_length
        self.y = torch.from_numpy(dataframe[target].to_numpy(dtype=np.float32)).float()
        self.X = torch.from_numpy(
            dataframe[features].to_numpy(dtype=np.float32)
        ).float()

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, i):
        if i >= self.sequence_length - 1:
            i_start = i - self.sequence_length + 1
            x = self.X[i_start : (i + 1), :]
        else:
            padding = self.X[0].repeat(self.sequence_length - i - 1, 1).float()
            x = self.X[0 : (i + 1), :]
            x = torch.cat((padding, x), 0).float()

        return x, self.y[i]


class LstmModel(postprocessing.FullPipeline, TimeSeriesDataset):
    """
    This class stores all model training and prediction methods for classification tasks.
    This class stores all pipeline relevant information (inherited from cpu preprocessing).
    The attribute "df_dict" always holds train and test as well as
    to predict data. The attribute "preprocess_decisions" stores encoders and other information generated during the
    model training. The attributes "predicted_classes" and "predicted_probs" store dictionaries (model names are dictionary keys)
    with predicted classes and probabilities (classification tasks) while "predicted_values" stores regression based
    predictions. The attribute "evaluation_scores" keeps track of model evaluation metrics (in dictionary format).
    :param datasource: Expects a Pandas dataframe (containing the target feature as a column)
    :param target_variable: Name of the target feature's column within the datasource dataframe.
    :param date_columns: Date columns can be passed as lists additionally for respective preprocessing. If not provided
    e2eml will try to detect datetime columns automatically. Date format is expected as YYYY-MM-DD anyway.
    :param categorical_columns: Categorical columns can be passed as lists additionally for respective preprocessing.
    If not provided e2eml will try to detect categorical columns automatically.
    :param nlp_columns: NLP columns expect a string declaring one text column.
    :param unique_identifier: A unique identifier (i.e. an ID column) can be passed as well to preserve this information
     for later processing.
    :param ml_task: Can be 'binary', 'multiclass' or 'regression'. On default will be determined automatically.
    :param preferred_training_mode: Must be 'cpu', if e2eml has been installed into an environment without LGBM and Xgboost on GPU.
    Can be set to 'gpu', if LGBM and Xgboost have been installed with GPU support. The default 'auto' will detect GPU support
    and optimize accordingly. (Default: 'auto')
    :param logging_file_path: Preferred location to save the log file. Will otherwise stored in the current folder.
    :param low_memory_mode: Adds a preprocessing feature to reduce dataframe memory footprint. Will lead to a loss in
    model performance. Will be extended by further memory savings features in future releases.
    However we highly recommend GPU usage to heavily decrease model training times.
    """

    def create_lstm_train_dataset(self):
        logging.info("Create LSTM train dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        train_dataset = TimeSeriesDataset(
            X_train[self.preprocess_decisions["n_features"]],
            self.target_variable,
            list(
                self.preprocess_decisions["n_features"].difference(
                    [self.target_variable]
                )
            ),
            sequence_length=self.lstm_settings["seq_len"],
        )
        return train_dataset

    def create_lstm_test_dataset(self):
        logging.info("Create LSTM test dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        test_dataset = TimeSeriesDataset(
            X_test[self.preprocess_decisions["n_features"]],
            self.target_variable,
            list(
                self.preprocess_decisions["n_features"].difference(
                    [self.target_variable]
                )
            ),
            sequence_length=self.lstm_settings["seq_len"],
        )
        return test_dataset

    def create_lstm_pred_dataset(self):
        logging.info("Create NLP prediction dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if self.prediction_mode:
            self.dataframe[self.target_variable] = 999  # creating dummy column
            # dummy_target = self.dataframe[self.target_variable]
            # self.dataframe = self.dataframe.drop(self.target_variable, axis=1)
            pred_dataset = TimeSeriesDataset(
                self.dataframe[self.preprocess_decisions["n_features"]],
                self.target_variable,
                list(
                    self.preprocess_decisions["n_features"].difference(
                        [self.target_variable]
                    )
                ),
                sequence_length=self.lstm_settings["seq_len"],
            )
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            pred_dataset = TimeSeriesDataset(
                X_test[self.preprocess_decisions["n_features"]],
                self.target_variable,
                list(
                    self.preprocess_decisions["n_features"].difference(
                        [self.target_variable]
                    )
                ),
                sequence_length=self.lstm_settings["seq_len"],
            )
        return pred_dataset

    def create_lstm_train_dataloader(self, train_batch_size=None, workers=None):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.lstm_settings["train_batch_size"]

        if workers:
            pass
        else:
            workers = self.lstm_settings["num_workers"]
        logging.info("Create Neural network train dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        train_dataset = self.create_lstm_train_dataset()
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            num_workers=workers,
            pin_memory=True,
        )
        return train_dataloader

    def create_lstm_test_dataloader(self, test_batch_size=None, workers=None):
        if test_batch_size:
            pass
        else:
            test_batch_size = self.lstm_settings["test_batch_size"]

        if workers:
            pass
        else:
            workers = self.lstm_settings["num_workers"]
        logging.info("Create LSTM test dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        test_dataset = self.create_lstm_test_dataset()
        test_dataloader = DataLoader(
            test_dataset,
            batch_size=test_batch_size,
            shuffle=False,
            num_workers=workers,
            pin_memory=True,
        )
        return test_dataloader

    def pred_lstm_dataloader(self, pred_batch_size=None, workers=None):
        if pred_batch_size:
            pass
        else:
            pred_batch_size = self.lstm_settings["pred_batch_size"]

        if workers:
            pass
        else:
            workers = self.lstm_settings["num_workers"]
        logging.info("Create Neural network prediction dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        pred_dataset = self.create_lstm_pred_dataset()
        pred_dataloader = DataLoader(
            pred_dataset,
            batch_size=pred_batch_size,
            shuffle=False,
            num_workers=workers,
            pin_memory=True,
        )
        return pred_dataloader

    def loss_fn(self, output, target):
        if self.lstm_settings["regression_loss"] == "mse":
            return torch.sqrt(nn.MSELoss()(output, target))
        elif self.lstm_settings["regression_loss"] == "smoothl1":
            return torch.sqrt(nn.SmoothL1Loss()(output, target))
        elif self.lstm_settings["regression_loss"] == "l1":
            return torch.sqrt(nn.L1Loss()(output, target))
        elif self.lstm_settings["regression_loss"] == "poisson":
            return torch.sqrt(nn.PoissonNLLLoss()(output, target))

    def get_num_features(self):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        N, D = X_train.shape
        self.preprocess_decisions["num_features"] = D
        return D

    def get_lstm_architecture(
        self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob
    ):
        class LSTMModel(nn.Module):
            def __init__(
                self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob
            ):
                super(LSTMModel, self).__init__()

                # Defining the number of layers and the nodes in each layer
                self.hidden_dim = hidden_dim
                self.layer_dim = layer_dim

                # LSTM layers
                self.lstm = nn.LSTM(
                    input_dim,
                    hidden_dim,
                    num_layers=layer_dim,
                    batch_first=True,
                    dropout=dropout_prob,
                )

                # Fully connected layer
                self.fc = nn.Linear(hidden_dim, output_dim)

            def forward(self, x):
                # Initializing hidden state for first input with zeros
                h0 = torch.zeros(
                    self.layer_dim, x.size(0), self.hidden_dim, dtype=torch.float32
                ).to(device)

                # Initializing cell state for first input with zeros
                c0 = torch.zeros(
                    self.layer_dim, x.size(0), self.hidden_dim, dtype=torch.float32
                ).to(device)

                # We need to detach as we are doing truncated backpropagation through time (BPTT)
                # If we don't, we'll backprop all the way to the start even after going through another batch
                # Forward propagation by passing in the input, hidden state, and cell state into the model
                # Propagate input through LSTM
                ula, (h_out, _) = self.lstm(x, (h0.detach(), c0.detach()))

                h_out = h_out[0].view(-1, self.hidden_dim)

                out = self.fc(h_out)

                return out

        model = LSTMModel(input_dim, hidden_dim, layer_dim, output_dim, dropout_prob)
        return model

    def lstm_model_setup(self, epochs=None):
        if self.prediction_mode:
            pass
        else:
            logging.info("Define LSTM model.")
            logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            self.get_num_features()
            model = self.get_lstm_architecture(
                input_dim=X_train[self.preprocess_decisions["n_features"]].shape[1]
                - 1,  # removes target from count
                hidden_dim=self.lstm_settings["hidden_dim"],
                layer_dim=self.lstm_settings["layer_dim"],
                output_dim=1,
                dropout_prob=self.lstm_settings["drop_out"],
            )
            model.to(device)
            model.train()
            LR = self.lstm_settings["learning_rate"]
            optimizer = torch.optim.Adam(
                model.parameters(),
                lr=LR,
                weight_decay=self.lstm_settings["weight_decay"],
            )

            if epochs:
                pass
            else:
                epochs = self.lstm_settings["epochs"]
            epochs = epochs
            train_steps = int(
                len(X_train) / self.lstm_settings["train_batch_size"] * epochs
            )
            num_steps = int(train_steps * 0.1)
            scheduler = get_linear_schedule_with_warmup(
                optimizer, num_steps, train_steps
            )
            return model, optimizer, train_steps, num_steps, scheduler

    def lstm_training(self, train_dataloader, model, optimizer, scheduler):
        logging.info("Start LSTM training loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        model.train()
        allpreds = []
        alltargets = []
        if self.shuffle_during_training:
            self.reset_test_train_index(drop_target=False)

        for X_train_batch, y_train_batch in train_dataloader:
            losses = []
            optimizer.zero_grad()

            with torch.cuda.amp.autocast():
                X_train_batch, y_train_batch = X_train_batch.to(
                    device
                ), y_train_batch.to(device)

                y_train_pred = model(X_train_batch)
                y_train_batch = y_train_batch.unsqueeze(1)
                loss = self.loss_fn(y_train_pred, y_train_batch)

                # For scoring
                losses.append(loss.item() / len(y_train_pred))
                allpreds.append(y_train_pred.detach().cpu().numpy())
                alltargets.append(y_train_batch.detach().squeeze(-1).cpu().numpy())

            scaler.scale(loss).backward()  # backwards of loss
            scaler.step(optimizer)  # Update optimizer
            scaler.update()  # scaler update
            scheduler.step()  # Update learning rate schedule

            # Combine dataloader minutes

        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        try:
            train_rme_loss = np.sqrt(
                mean_squared_error(
                    np.array(alltargets), np.array(allpreds, dtype="float64")
                )
            )
        except ValueError:
            all_preds_arr = np.array(allpreds)
            all_preds_arr[np.isnan(all_preds_arr)] = 0
            all_preds_arr[all_preds_arr == -inf] = 0
            all_preds_arr[all_preds_arr == inf] = 0
            train_rme_loss = np.sqrt(
                mean_squared_error(np.array(alltargets), all_preds_arr)
            )
        return losses, train_rme_loss

    def lstm_validating(self, valid_dataloader, model):
        logging.info("Start NLP validation loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        model.eval()
        allpreds = []
        alltargets = []
        if self.shuffle_during_training:
            self.reset_test_train_index(drop_target=False)

        for X_test_batch, y_test_batch in valid_dataloader:
            losses = []

            with torch.no_grad():
                X_test_batch, y_test_batch = X_test_batch.to(device), y_test_batch.to(
                    device
                )

                y_train_pred = model(X_test_batch)

                y_test_batch = y_test_batch.unsqueeze(1)
                loss = self.loss_fn(y_train_pred, y_test_batch)

                # For scoring
                losses.append(loss.item() / len(y_train_pred))
                allpreds.append(y_train_pred.detach().cpu().numpy())
                alltargets.append(y_test_batch.detach().squeeze(-1).cpu().numpy())
                # Combine dataloader minutes

        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        try:
            valid_rme_loss = np.sqrt(
                mean_squared_error(
                    np.array(alltargets), np.array(allpreds, dtype="float64")
                )
            )
        except ValueError:
            all_preds_arr = np.array(allpreds)
            all_preds_arr[np.isnan(all_preds_arr)] = 0
            all_preds_arr[all_preds_arr == -inf] = 0
            all_preds_arr[all_preds_arr == inf] = 0
            valid_rme_loss = np.sqrt(
                mean_squared_error(np.array(alltargets), all_preds_arr)
            )

        return allpreds, losses, valid_rme_loss

    def lstm_predicting(self, pred_dataloader, model, pathes):
        logging.info("Start Neural network prediction loop.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        allpreds = []
        model_no = 0
        mode_cols = []
        for m_path in pathes:
            state = torch.load(m_path)
            model.load_state_dict(state["state_dict"])
            model.to(device)
            model.eval()
            preds = []
            # allvalloss = 0
            with torch.no_grad():
                for X_pred_batch, y_pred_batch in pred_dataloader:
                    X_pred_batch, y_pred_batch = X_pred_batch.to(
                        device
                    ), y_pred_batch.to(device)
                    y_pred_pred = model(X_pred_batch)
                    y_pred_pred = y_pred_pred.squeeze(-1)
                    preds.append(y_pred_pred.detach().cpu().numpy())

                preds = np.concatenate(preds)

                if self.prediction_mode:
                    self.dataframe[f"preds_model{model_no}"] = preds
                else:
                    X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
                    X_test[f"preds_model{model_no}"] = preds
                mode_cols.append(f"preds_model{model_no}")

                allpreds.append(preds)
                model_no += 1
            del state
            torch.cuda.empty_cache()
            _ = gc.collect()
        return allpreds, mode_cols

    def load_model_states(self, path=None):
        logging.info("Load model save states.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if path:
            pass
        else:
            path = os.getcwd()
        if self.prediction_mode:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if ".pth" in s
            ]
            return pthes
        else:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if ".pth" in s
            ]
            return pthes

    def lstm_train(self):
        if self.prediction_mode:
            pass
        else:
            logging.info("Start LSTM training.")
            logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
            if self.shuffle_during_training:
                self.reset_test_train_index(drop_target=False)

            train_dataloader = self.create_lstm_train_dataloader()
            test_dataloader = self.create_lstm_test_dataloader()
            (
                model,
                optimizer,
                train_steps,
                num_steps,
                scheduler,
            ) = self.lstm_model_setup()
            scheduler = get_linear_schedule_with_warmup(
                optimizer, num_steps, train_steps
            )

            trainlosses = []
            vallosses = []
            bestscore = None
            trainscores = []
            validscores = []

            for epoch in tqdm(range(self.lstm_settings["epochs"])):
                print("---------------" + str(epoch) + "start-------------")
                trainloss, trainscore = self.lstm_training(
                    train_dataloader, model, optimizer, scheduler
                )
                trainlosses.append(trainloss)
                trainscores.append(trainscore)
                print("trainscore is " + str(trainscore))
                preds, validloss, valscore = self.lstm_validating(
                    test_dataloader, model
                )
                vallosses.append(validloss)
                validscores.append(valscore)

                print("valscore is " + str(valscore))
                if bestscore is None:
                    bestscore = valscore
                    print("Save first model")
                    state = {
                        "state_dict": model.state_dict(),
                        "optimizer_dict": optimizer.state_dict(),
                        "bestscore": bestscore,
                    }
                    torch.save(state, "model0.pth")

                elif bestscore > valscore:
                    bestscore = valscore
                    print("found better point")
                    state = {
                        "state_dict": model.state_dict(),
                        "optimizer_dict": optimizer.state_dict(),
                        "bestscore": bestscore,
                    }
                    torch.save(state, "model0.pth")
                else:
                    pass

            bestscores = []
            bestscores.append(bestscore)

            for fold in range(self.lstm_settings["nb_model_to_create"]):

                if self.shuffle_during_training:
                    self.reset_test_train_index(drop_target=False)
                X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

                # initializing the data
                train_dataloader = self.create_lstm_train_dataloader()
                test_dataloader = self.create_lstm_test_dataloader()

                model = self.get_lstm_architecture(
                    input_dim=X_train[self.preprocess_decisions["n_features"]].shape[1]
                    - 1,  # removes target from count
                    hidden_dim=self.lstm_settings["hidden_dim"],
                    layer_dim=self.lstm_settings["layer_dim"],
                    output_dim=1,
                    dropout_prob=self.lstm_settings["drop_out"],
                )
                model.to(device)
                LR = self.lstm_settings["learning_rate"]
                optimizer = AdamW(
                    model.parameters(), LR, betas=(0.99, 0.999), weight_decay=1e-2
                )
                train_steps = int(
                    len(X_train)
                    / self.lstm_settings["train_batch_size"]
                    * self.lstm_settings["epochs"]
                )
                num_steps = int(train_steps * 0.1)
                scheduler = get_linear_schedule_with_warmup(
                    optimizer, num_steps, train_steps
                )

                trainlosses = []
                vallosses = []
                bestscore = None
                trainscores = []
                validscores = []

                for epoch in tqdm(range(self.lstm_settings["epochs"])):
                    print("---------------" + str(epoch) + "start-------------")
                    trainloss, trainscore = self.lstm_training(
                        train_dataloader, model, optimizer, scheduler
                    )
                    trainlosses.append(trainloss)
                    trainscores.append(trainscore)

                    print("trainscore is " + str(trainscore))
                    preds, validloss, valscore = self.lstm_validating(
                        test_dataloader, model
                    )

                    vallosses.append(validloss)
                    validscores.append(valscore)
                    print("valscore is " + str(valscore))

                    if bestscore is None:
                        bestscore = valscore
                        print("Save first model")
                        state = {
                            "state_dict": model.state_dict(),
                            "optimizer_dict": optimizer.state_dict(),
                            "bestscore": bestscore,
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    elif bestscore > valscore:
                        bestscore = valscore
                        print("found better point")
                        state = {
                            "state_dict": model.state_dict(),
                            "optimizer_dict": optimizer.state_dict(),
                            "bestscore": bestscore,
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    else:
                        pass
                bestscores.append(bestscore)

            del model, optimizer, scheduler
            _ = gc.collect()

    def lstm_predict(self):
        logging.info("Start LSTM transformer prediction.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if self.shuffle_during_training:
            self.reset_test_train_index(drop_target=False)
        if self.prediction_mode:
            model = self.get_lstm_architecture(
                input_dim=self.dataframe[self.preprocess_decisions["n_features"]].shape[
                    1
                ]
                - 1,  # removes target from count
                hidden_dim=self.lstm_settings["hidden_dim"],
                layer_dim=self.lstm_settings["layer_dim"],
                output_dim=1,
                dropout_prob=self.lstm_settings["drop_out"],
            )
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            model = self.get_lstm_architecture(
                input_dim=X_test[self.preprocess_decisions["n_features"]].shape[1]
                - 1,  # removes target from count
                hidden_dim=self.lstm_settings["hidden_dim"],
                layer_dim=self.lstm_settings["layer_dim"],
                output_dim=1,
                dropout_prob=self.lstm_settings["drop_out"],
            )
        pathes = self.load_model_states(path=self.tabular_nn_model_save_states_path)
        pthes = self.load_model_states(path=self.tabular_nn_model_save_states_path)
        try:
            for path in pathes:
                if re.search("generator_model.pth", path):
                    pthes.remove(path)
                if re.search("autoencoder_model.pth", path):
                    pthes.remove(path)
        except Exception:
            pass
        print(pthes)
        pred_dataloader = self.pred_lstm_dataloader()
        allpreds, mode_cols = self.lstm_predicting(pred_dataloader, model, pthes)

        if self.prediction_mode:
            self.dataframe["lstm_median"] = self.dataframe[mode_cols].median(axis=1)
            self.dataframe["lstm_mean"] = self.dataframe[mode_cols].mean(axis=1)

            # reverse scale target and df
            self.dataframe[self.target_variable] = self.dataframe["lstm_mean"]
            self.scale_with_target(mode="reverse", drop_target=False)

            self.dataframe[self.target_variable] = self.target_skewness_handling(
                preds_to_reconvert=self.dataframe[self.target_variable].values,
                mode="revert",
            )
            self.predicted_values["lstm"] = self.dataframe[self.target_variable]

        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

            if self.lstm_settings["keep_best_model_only"]:
                # we check, if one savestate underperforms and delete him out
                scorings = []
                states = []
                for state in mode_cols:
                    state_score = self.median_abs_error_eval(Y_test, X_test[state])
                    # print(state_score)
                    scorings.append(state_score)
                    states.append(state)
                scorings_arr = np.array(scorings)
                scorings_mean = np.mean(scorings_arr)
                scorings_std = np.std(scorings_arr)
                # print(scorings_std)
                keep_state = scorings_arr < scorings_mean + scorings_std
                for index, state in enumerate(states):
                    os_string = state[-6:]
                    if keep_state.tolist()[index]:
                        pass
                    else:
                        states.remove(state)
                        X_test.drop(state, axis=1)
                        os.remove(f"{os_string}.pth")
            else:
                pass

            X_test["lstm_median"] = X_test[mode_cols].median(axis=1)
            X_test["lstm_mean"] = X_test[mode_cols].mean(axis=1)
            # reverse scale target and df
            X_test[self.target_variable] = X_test["lstm_mean"]
            self.scale_with_target(mode="reverse", drop_target=False)

            X_test[self.target_variable] = self.target_skewness_handling(
                preds_to_reconvert=X_test[self.target_variable].values, mode="revert"
            )
            self.predicted_values["lstm"] = X_test[self.target_variable]
            self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)



import numpy as np
import optuna
import pandas as pd


class AutoDataSlicer:
    def __init__(
        self,
        df,
        target_col,
        objective="regression",
        use_percentages_as_metric=True,
        study_direction="minimize",
        study_name="dataframe slicer",
        random_state=1000,
        n_trials=None,
        remove_high_cardinality=True,
        max_cat_col_cardinality=50,
        supress_granular_prints=True,
        slice_one_category_only=True,
        pre_remove_slices=False,
        timeout=3 * 60 * 60,
    ):
        self.df = df
        self.target_col = target_col
        self.objective = objective
        self.use_percentages_as_metric = use_percentages_as_metric
        self.study_direction = study_direction
        self.study_name = study_name
        self.random_state = random_state
        if isinstance(n_trials, list):
            self.n_trials = n_trials
        else:
            self.n_trials = [500, 100, 100]
        self.remove_high_cardinality = remove_high_cardinality
        self.max_cat_col_cardinality = max_cat_col_cardinality
        self.supress_granular_prints = supress_granular_prints
        self.timeout = timeout
        self.slice_one_category_only = slice_one_category_only
        self.pre_remove_slices = pre_remove_slices
        self.study = None
        self.cat_columns = None
        self.num_col_list = None
        self.numerical_meta_data = None
        self.categorical_filters = {}
        self.trial_counter = 0
        self.best_segments = None
        self.best_score = 1

    def get_categorical_columns(self):
        self.cat_columns = self.df.select_dtypes(include=["object"]).columns.to_list()
        if self.n_trials[0] > 0 and len(self.cat_columns) == 0:
            self.n_trials[0] = 0
            self.n_trials[2] = 0
            print(
                "Categorical columns could not be identified. Corrected n_trials attribute."
            )

    def get_numerical_columns(self):
        num_dtypes = [
            "int8",
            "int16",
            "int32",
            "int64",
            "float16",
            "float32",
            "float64",
        ]

        self.num_col_list = []
        for vartype in num_dtypes:
            num_cols = self.df.select_dtypes(include=[vartype]).columns
            for col in num_cols:
                self.num_col_list.append(col)

        if self.n_trials[1] > 0 and len(self.num_col_list) == 0:
            self.n_trials[1] = 0
            self.n_trials[2] = 0
            print(
                "Numerical columns could not be identified. Corrected n_trials attribute."
            )

    def get_numerical_meta_data(self):
        self.numerical_meta_data = self.df[self.num_col_list].describe()

    def create_categorical_filters(self):
        for col in self.cat_columns:
            self.categorical_filters[col] = {}
            options = self.df[col].unique()
            for option in options:
                self.categorical_filters[col][option] = 1

    def remove_cardinal_cols(self):
        uniques = self.df[self.cat_columns].nunique()
        for col in self.cat_columns:
            if uniques[col] > self.max_cat_col_cardinality:
                self.cat_columns.remove(col)
                self.df = self.df.drop(col, axis=1)
                print(f"Removed column {col} due to high cardinality.")

    def pre_slice_dataframe(self):
        # build slices based on categories
        for col, val_list in self.categorical_filters.items():  # noqa: B007
            for cat, value in self.categorical_filters[col].items():  # noqa: B007
                temp_df = self.df[(self.df[col] == cat)]
                try:
                    has_true = (
                        temp_df[self.target_col]
                        .value_counts(normalize=self.use_percentages_as_metric)
                        .loc[True]
                    )
                    # exclude each category that does not include any representatives of target class
                    if not has_true:
                        self.df = self.df[(self.df[col] != cat)].copy()
                        print(
                            f"Excluded from feature {col} sub category {cat} due to missing data points of target class."
                        )
                except KeyError:
                    self.df = self.df[(self.df[col] != cat)].copy()
                    print(
                        f"Excluded from feature {col} sub category {cat} due to missing data points of target class."
                    )

    def pre_slice_numerical_dataframe(self):
        for col in self.num_col_list:
            perc_25th = self.numerical_meta_data[col]["25%"]
            perc_75th = self.numerical_meta_data[col]["75%"]
            temp_df = self.df[(self.df[col] < perc_25th)]
            try:
                has_true = (
                    temp_df[self.target_col]
                    .value_counts(normalize=self.use_percentages_as_metric)
                    .loc[True]
                )
                # exclude each category that does not include any representatives of target class
                if not has_true:
                    self.df = self.df[(self.df[col] >= perc_25th)].copy()
                    print(
                        f"Excluded from feature {col} values below {perc_25th} due to missing data points of target class."
                    )
            except KeyError:
                self.df = self.df[(self.df[col] >= perc_25th)].copy()
                print(
                    f"Excluded from feature {col} values below {perc_25th} due to missing data points of target class."
                )

            temp_df = self.df[(self.df[col] > perc_75th)]
            try:
                has_true = (
                    temp_df[self.target_col]
                    .value_counts(normalize=self.use_percentages_as_metric)
                    .loc[True]
                )
                # exclude each category that does not include any representatives of target class
                if not has_true:
                    self.df = self.df[(self.df[col] <= perc_75th)].copy()
                    print(
                        f"Excluded from feature {col} values above {perc_75th} due to missing data points of target class."
                    )
            except KeyError:
                self.df = self.df[(self.df[col] <= perc_75th)].copy()
                print(
                    f"Excluded from feature {col} values above {perc_75th} due to missing data points of target class."
                )

    def append_results(self, segment_dict, numerical_segment_dict, score, df_len, mode):
        if isinstance(self.best_segments, pd.DataFrame):
            new_res = pd.DataFrame(
                {
                    "score": score,
                    "segments_chosen": [segment_dict],
                    "numerical_segments_chosen": [numerical_segment_dict],
                    "no_rows_considered": df_len,
                    "mode": mode,
                }
            )
            self.best_segments = pd.concat([self.best_segments, new_res])
            self.best_segments = self.best_segments.sort_values(
                by=["score"], ascending=[True]
            )
        else:
            self.best_segments = pd.DataFrame(
                {
                    "score": score,
                    "segments_chosen": [segment_dict],
                    "numerical_segments_chosen": [numerical_segment_dict],
                    "no_rows_considered": df_len,
                    "mode": mode,
                }
            )

    def get_best_slice(self, min_rows_considered=20):
        best_df = self.best_segments
        best_df = best_df[(best_df["no_rows_considered"] >= min_rows_considered)]
        mode = best_df.head(1)["mode"].values[0]
        temp_df = self.df.copy()

        if mode in ["Categorical only", "Categorical and numerical"]:
            top_cat_slice_param = best_df["segments_chosen"]
            for col, val_list in (top_cat_slice_param.values[0]).items():  # noqa: B007
                for cat, value in (
                    top_cat_slice_param.values[0][col]
                ).items():  # noqa: B007
                    if value == 0:
                        temp_df = temp_df[(temp_df[col] != cat)]
                    else:
                        pass

        if mode in ["Numerical only", "Categorical and numerical"]:
            top_num_slice_param = best_df.head(1)["numerical_segments_chosen"]
            col = top_num_slice_param.values[0]["chosen_col"]
            min_val = top_num_slice_param.values[0][col]["min"]
            max_val = top_num_slice_param.values[0][col]["max"]
            temp_df = temp_df[(temp_df[col] >= min_val) & (temp_df[col] < max_val)]

        return temp_df

    def slice_df_binary(self):  # noqa: C901
        def binary_objective(trial):  # this does not allow additional parameters!

            self.trial_counter += 1
            print(f"Start trial {self.trial_counter}...")

            if self.trial_counter <= self.n_trials[0]:
                mode = "Categorical only"
            elif self.trial_counter > self.n_trials[0] and self.trial_counter <= np.sum(
                np.array(self.n_trials[:2])
            ):
                mode = "Numerical only"
            elif self.trial_counter > np.sum(np.array(self.n_trials[:2])):
                mode = "Categorical and numerical"

            if self.trial_counter <= self.n_trials[0] or self.trial_counter > np.sum(
                np.array(self.n_trials[:2])
            ):
                # store all slice choices
                param = {}

                if self.slice_one_category_only:
                    col = trial.suggest_categorical("chosen_cat_col", self.cat_columns)
                    param[col] = {}
                    for cat, value in self.categorical_filters[  # noqa: B007
                        col
                    ].items():  # noqa: B007
                        param[col][cat] = trial.suggest_int(f"{col}_{cat}", 0, 1)

                else:

                    # build slices based on categories
                    for col, val_list in self.categorical_filters.items():  # noqa: B007
                        param[col] = {}
                        for cat, value in self.categorical_filters[  # noqa: B007
                            col
                        ].items():  # noqa: B007
                            param[col][cat] = trial.suggest_int(f"{col}_{cat}", 0, 1)

                temp_df = self.df.copy()

                # slice df based on selected filters
                for col, val_list in param.items():  # noqa: B007
                    for cat, value in param[col].items():  # noqa: B007
                        if value == 0:
                            temp_df = temp_df[(temp_df[col] != cat)]
                        else:
                            pass

            if self.trial_counter > self.n_trials[0] and self.trial_counter <= np.sum(
                np.array(self.n_trials[:2])
            ):
                temp_df = self.df.copy()
                param = {}

            if self.trial_counter > self.n_trials[0]:
                # here starts sclicing of numerical columns
                numerical_param = {}
                numerical_param["chosen_col"] = trial.suggest_categorical(
                    "chosen_col", self.num_col_list
                )

                col = numerical_param["chosen_col"]
                numerical_param[col] = {}
                min_val = self.numerical_meta_data[col]["min"]
                perc_75th = self.numerical_meta_data[col]["75%"]
                max_val = self.numerical_meta_data[col]["max"]
                numerical_param[col]["min"] = trial.suggest_uniform(
                    f"{col}_min", min_val, perc_75th
                )
                numerical_param[col]["max"] = trial.suggest_uniform(
                    f"{col}_max", numerical_param[col]["min"], max_val
                )

                # slice df based on selected filters
                min_val = numerical_param[col]["min"]
                max_val = numerical_param[col]["max"]
                temp_df = temp_df[(temp_df[col] >= min_val) & (temp_df[col] < max_val)]
            else:
                numerical_param = {}

            df_len = len(temp_df.index)

            try:
                # print(temp_df[self.target_col].value_counts(normalize=self.use_percentages_as_metric))
                mae = temp_df[self.target_col].sum() / df_len
                self.append_results(param, numerical_param, mae, df_len, mode)
            except KeyError:
                mae = 999999999

            if self.supress_granular_prints:
                print(
                    f"""Trial {self.trial_counter} finished with score {mae} with df length of {df_len}."""
                )

            if self.trial_counter == 0 or mae < self.best_score:
                self.best_score = mae

            del temp_df
            return mae, df_len

        if self.supress_granular_prints:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

        # create filters
        self.get_categorical_columns()

        if self.remove_high_cardinality:
            self.remove_cardinal_cols()
            self.get_categorical_columns()
        self.create_categorical_filters()

        if self.pre_remove_slices:
            self.pre_slice_dataframe()
        self.get_numerical_columns()
        self.get_numerical_meta_data()

        if self.pre_remove_slices:
            self.pre_slice_numerical_dataframe()
            self.get_numerical_meta_data()

        sampler = optuna.samplers.TPESampler(
            multivariate=True, seed=self.random_state, warn_independent_sampling=False
        )
        study = optuna.create_study(
            directions=[self.study_direction, "maximize"],
            sampler=sampler,
            study_name=self.study_name,
        )
        study.optimize(
            binary_objective,
            n_trials=np.sum(np.array(self.n_trials)),
            gc_after_trial=True,
            show_progress_bar=True,
            timeout=self.timeout,
        )
        self.study = study

    def plot_study_optimization_history(self, target=lambda t: t.values[0]):
        try:
            fig = optuna.visualization.plot_optimization_history(
                self.study, target=target
            )
            fig.show()
        except ZeroDivisionError:
            pass

    def plot_study_parameter_importances(self, target=lambda t: t.values[0]):
        try:
            fig = optuna.visualization.plot_param_importances(self.study, target=target)
            fig.show()
        except ZeroDivisionError:
            pass


def analyse_results(
    holdout_df,
    y_true,
    y_hat,
    study_direction="minimize",
    n_trials=None,
    random_state=1000,
    max_cat_col_cardinality=50,
):
    """
    Takes a Pandas DataFrame and tries to find slices with high quotas of
    search_for param. During optimization it will try to optimize for slices as big as possible as well.
    If only a certain class is of interest, just pass the slice containing this class.
    :param holdout_df: Pandas DataFrame without target or prediction column.
    :param y_true: Pandas Series holding real target values. Index must match holdout_df
    :param y_hat: Pandas Series holding predicted target values. Index must match holdout_df
    :param study_direction: Expects either 'maximize' or 'minimize'. According to this the routine
    will try to find segments with maximum or minimum error rates.
    :param n_trials: Expects a list containing three integers to define the number of trials for each part of the
    routine (default: [500, 100, 100]). Index 0 will slice using categories only. Index 1 will slice numerical
    columns only. Index 2 will slice categorical and numerical columns.
    :param random_state: Integer specifying random state to use.
    :param max_cat_col_cardinality: Expects integer. Columns with specified cardinality or higher will be removed.
    :return: Returns AutoDataSlicer object.
    """
    holdout_df["true_target"] = y_true
    holdout_df["pred_target"] = y_hat
    holdout_df["abs_delta"] = np.abs(
        holdout_df["true_target"] - holdout_df["pred_target"]
    )

    study_maker_fl = AutoDataSlicer(
        holdout_df,
        target_col="abs_delta",
        objective="regression",
        study_direction=study_direction,
        study_name="dataframe slicer",
        random_state=random_state,
        n_trials=n_trials,
        remove_high_cardinality=True,
        max_cat_col_cardinality=max_cat_col_cardinality,
    )
    study_maker_fl.slice_df_binary()
    return study_maker_fl



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from statsmodels.tsa.stattools import adfuller, kpss

from e2eml.full_processing import cpu_preprocessing

plt.style.use("fivethirtyeight")


class TimeSeriesPreprocessing(cpu_preprocessing.PreProcessing):
    def time_series_unpacking(self):
        """
        Takes the datasource dictionary within the class and unpacks it, depending on the provided source format.
        :return: Returns unpacked data objects.
        """
        if self.source_format == "numpy array":
            Y_train, Y_test = self.np_array_unpack_test_train_dict()
            return Y_train, Y_test
        elif self.source_format == "Pandas dataframe":
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            return X_train, X_test, Y_train, Y_test

    def keep_target_only(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            X_train = X_train[[self.target_variable]]
            X_test = X_test[[self.target_variable]]
            self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def reattach_targets(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            X_train[self.target_variable] = Y_train
            X_test[self.target_variable] = Y_test
            self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def sliding_windows(self, data, seq_length=5):
        x = []
        y = []

        for i in range(len(data) - seq_length - 1):
            _x = data[i : (i + seq_length)]
            _y = data[i + seq_length]
            x.append(_x)
            y.append(_y)

        return np.array(x), np.array(y)

    def create_sliding_windows(self):
        x, y = self.sliding_windows(self.dataframe.to_numpy())
        train_size = int(len(y) * 0.30)

        X_train = np.array(x[0:train_size])
        Y_train = np.array(y[0:train_size])

        X_test = np.array(x[train_size : len(x)])
        Y_test = np.array(y[train_size : len(y)])
        self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def add_rolling_feature(self, dataframe, col_list, marker_text, time_window):
        """Takes a dataframe, column to use and a time window to create a rolling mean.
        :params:
        dataframe: the dataframe to use the feature from
        col:_name: the column to be used for the rolling feature
        col_list: list of calls to loop through
        marker_text: unique marker to prevent join attachments like x & y
        time_window: number of days to roll over
        returns: the modified dataframe
        """
        for columns in col_list:
            dataframe[f"rolling_avg_{columns}_{time_window}D_{marker_text}"] = (
                dataframe[columns].rolling(window=time_window).mean()
            )
        return dataframe

    def time_series_wrap_test_train_to_dict(
        self, X_train=None, X_test=None, Y_train=None, Y_test=None
    ):
        """
        Takes either four Pandas dataframes and series or two numpy arrays and packs them into the class
        dictionary.
        :param X_train: Pandas dataframe (optional)
        :param X_test: Pandas dataframe (optional)
        :param Y_train: Pandas series or numpy array
        :param Y_test: Pandas series or numpy array
        :return: Updates class and return dictionary.
        """
        if self.source_format == "numpy array":
            return self.np_array_wrap_test_train_to_dict(Y_train, Y_test)
        elif self.source_format == "Pandas dataframe":
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def augmented_dickey_fuller_test(self, window):
        """
        The Dickey Fuller test is one of the most popular statistical tests. It can be used to determine the
        presence of unit root in the series, and hence help us understand if the series is stationary or not.
        The null and alternate hypothesis of this test are:
          - Null Hypothesis: The series has a unit root (value of a =1)
          - Alternate Hypothesis: The series has no unit root.
        If we fail to reject the null hypothesis, we can say that the series is non-stationary.
        This means that the series can be linear or difference stationary.
        :return:
        """
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        if self.auto_arima_settings["plot_adf_kpss"]:
            # Determing rolling statistics
            rolmean = X_train.rolling(window=window).mean()
            rolstd = X_train.rolling(window=window).std()
            # Plot rolling statistics:
            plt.plot(X_train, color="blue", label="Original")
            plt.plot(rolmean, color="red", label="Rolling Mean")
            plt.plot(rolstd, color="black", label="Rolling Std")
            plt.legend(loc="best")
            plt.title("Rolling Mean & Standard Deviation")
            plt.show()
            # Perform Dickey-Fuller test:
            print("Results of Dickey-Fuller Test:")
        dftest = adfuller(X_train, autolag="AIC")
        dfoutput = pd.Series(
            dftest[0:4],
            index=[
                "Test Statistic",
                "p-value",
                "#Lags Used",
                "Number of Observations Used",
            ],
        )
        for key, value in dftest[4].items():
            dfoutput["Critical Value (%s)" % key] = value
        if self.auto_arima_settings["plot_adf_kpss"]:
            print(dfoutput)
        if dfoutput["p-value"] >= 0.5:
            self.preprocess_decisions["adf_test_is_stationary"] = False
        else:
            self.preprocess_decisions["adf_test_is_stationary"] = True

    def kwiatkowski_phillips_schmidt_shin_test(self):
        """
        KPSS is another test for checking the stationarity of a time series (slightly less popular than the Dickey
        Fuller test). The null and alternate hypothesis for the KPSS test are opposite that of the ADF test,
        which often creates confusion.
        The authors of the KPSS test have defined the null hypothesis as the process is trend stationary,
        to an alternate hypothesis of a unit root series.
        - Null Hypothesis: The process is trend stationary.
        - Alternate Hypothesis: The series has a unit root (series is not stationary).
        Test for stationarity: If the test statistic is greater than the critical value,
        we reject the null hypothesis (series is not stationary). If the test statistic is less than the critical value,
         if fail to reject the null hypothesis (series is stationary).
        So in summary, the ADF test has an alternate hypothesis of linear or difference stationary,
        while the KPSS test identifies trend-stationarity in a series.
        :return:
        """
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        try:
            kpsstest = kpss(X_train.values, regression="c")
            kpss_output = pd.Series(
                kpsstest[0:3], index=["Test Statistic", "p-value", "Lags Used"]
            )
            for key, value in kpsstest[3].items():
                kpss_output["Critical Value (%s)" % key] = value

            if self.auto_arima_settings["plot_adf_kpss"]:
                print("Results of KPSS Test:")
                print(kpss_output)
            if kpss_output["p-value"] < 0.5:
                self.preprocess_decisions["kpss_test_is_stationary"] = False
            else:
                self.preprocess_decisions["kpss_test_is_stationary"] = True
        except ValueError:
            kpss_output = {}
            kpss_output["p-value"] = "nan"
            if self.auto_arima_settings["plot_adf_kpss"]:
                print("KPSS test failed. Use ADF results only.")
            self.preprocess_decisions[
                "kpss_test_is_stationary"
            ] = self.preprocess_decisions["adf_test_is_stationary"]

    def make_stationary(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            self.preprocess_decisions["arima_nb_differentiations"] = 0
            self.preprocess_decisions["arima_transformation_used"] = "None"
            X_train_copy = X_train.copy()

            self.augmented_dickey_fuller_test(window=6)
            self.kwiatkowski_phillips_schmidt_shin_test()

            while (
                not self.preprocess_decisions["adf_test_is_stationary"]
                and not self.preprocess_decisions["kpss_test_is_stationary"]
            ):
                self.preprocess_decisions["arima_nb_differentiations"] += 1
                X_train = X_train.diff(
                    self.preprocess_decisions["arima_nb_differentiations"]
                ).dropna()
                self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)
                X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

                if (
                    self.preprocess_decisions["arima_nb_differentiations"]
                    > self.auto_arima_settings["max_d"]
                ):
                    X_train = X_train_copy
                    X_train = np.log1p(X_train)
                    X_train[np.isfinite(X_train) is False] = 0
                    Y_train = X_train
                    self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)
                    self.preprocess_decisions["arima_transformation_used"] = "log"

                self.augmented_dickey_fuller_test(window=6)
                self.kwiatkowski_phillips_schmidt_shin_test()

            Y_train = X_train

            if self.preprocess_decisions["arima_transformation_used"] == "log":
                Y_test = np.log1p(Y_test)
                Y_test[np.isfinite(Y_test) is False] = 0
            elif self.preprocess_decisions["arima_nb_differentiations"] > 0:
                length = len(Y_test.index)
                full_test = pd.concat([Y_train, Y_test])
                full_test = full_test.diff(
                    self.preprocess_decisions["arima_nb_differentiations"]
                )  # .dropna()
                full_test = full_test.ffill()
                Y_test = full_test.tail(length)
                Y_test = pd.Series(Y_test[self.target_variable].values)
            else:
                pass

            Y_train = pd.Series(Y_train[self.target_variable].values)
            self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def stationarity_explainer(self):
        explanation = """
        What are the different types of stationarity?
        - Strict Stationary: A strict stationary series satisfies the mathematical definition of a stationary process.
          For a strict stationary series, the mean, variance and covariance are not the function of time.
          The aim is to convert a non-stationary series into a strict stationary series for making predictions.
        - Trend Stationary: A series that has no unit root but exhibits a trend is referred to as a trend stationary
          series. Once the trend is removed, the resulting series will be strict stationary. The KPSS test classifies a
          series as stationary on the absence of unit root. This means that the series can be strict stationary or
          trend stationary.
        - Difference Stationary: A time series that can be made strict stationary by differencing falls under difference
          stationary. ADF test is also known as a difference stationarity test.

        It’s always better to apply both the tests, so that we are sure that the series is truly stationary.
        Let us look at the possible outcomes of applying these stationary tests.

        - Case 1: Both tests conclude that the series is not stationary -> series is not stationary
        - Case 2: Both tests conclude that the series is stationary -> series is stationary
        - Case 3: KPSS = stationary and ADF = not stationary -> trend stationary, remove the trend to make series
          strict stationary
        - Case 4: KPSS = not stationary and ADF = stationary -> difference stationary, use differencing to make series
          stationary
        """
        print(explanation)



import gc
import logging
import os
import re

# import lightgbm as lgb
import numpy as np
import pandas as pd
import psutil
import torch
import torch.nn as nn
import torch.optim as optim

# from IPython.display import clear_output
# from sklearn.metrics import matthews_corrcoef
from torch.autograd.variable import Variable
from torch.utils.data import DataLoader, Dataset

from e2eml.full_processing.postprocessing import FullPipeline

# from tqdm import tqdm
# from transformers import AdamW, get_linear_schedule_with_warmup


# specify GPU
scaler = torch.cuda.amp.GradScaler()  # GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class GanDataset(Dataset):
    def __init__(self, X_data, y_data):
        self.X_data = X_data
        self.y_data = y_data

    def __getitem__(self, index):
        return self.X_data[index], self.y_data[index]

    def __len__(self):
        return len(self.X_data)


class TabularGeneratorClassification(FullPipeline, GanDataset):
    def get_generator_discriminator(self):
        class Generator(nn.Module):
            """
            Creates fake data trying to trick the Discriminator.
            """

            def __init__(
                self,
                num_features,
                output_dim=1,
                sign_size=32,
                cha_input=16,
                cha_hidden=32,
                K=2,
                dropout_input=0.2,
                dropout_hidden=0.2,
                dropout_output=0.2,
            ):
                super().__init__()

                hidden_size = sign_size * cha_input
                sign_size1 = sign_size
                sign_size2 = sign_size // 2
                output_size = (sign_size // 4) * cha_hidden

                self.hidden_size = hidden_size
                self.cha_input = cha_input
                self.cha_hidden = cha_hidden
                self.K = K
                self.sign_size1 = sign_size1
                self.sign_size2 = sign_size2
                self.output_size = output_size
                self.dropout_input = dropout_input
                self.dropout_hidden = dropout_hidden
                self.dropout_output = dropout_output

                self.batch_norm1 = nn.BatchNorm1d(num_features)
                self.dropout1 = nn.Dropout(dropout_input)
                dense1 = nn.Linear(num_features, hidden_size, bias=False)
                self.dense1 = nn.utils.weight_norm(dense1)

                # 1st conv layer
                self.batch_norm_c1 = nn.BatchNorm1d(cha_input)
                conv1 = nn.Conv1d(
                    cha_input,
                    cha_input * K,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_input,
                    bias=False,
                )
                self.conv1 = nn.utils.weight_norm(conv1, dim=None)

                self.ave_po_c1 = nn.AdaptiveAvgPool1d(output_size=sign_size2)

                # 2nd conv layer
                self.batch_norm_c2 = nn.BatchNorm1d(cha_input * K)
                self.dropout_c2 = nn.Dropout(dropout_hidden)
                conv2 = nn.Conv1d(
                    cha_input * K,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv2 = nn.utils.weight_norm(conv2, dim=None)

                # 3rd conv layer
                self.batch_norm_c3 = nn.BatchNorm1d(cha_hidden)
                self.dropout_c3 = nn.Dropout(dropout_hidden)
                conv3 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                )
                self.conv3 = nn.utils.weight_norm(conv3, dim=None)

                # 4th conv layer
                self.batch_norm_c4 = nn.BatchNorm1d(cha_hidden)
                conv4 = nn.Conv1d(
                    cha_hidden,
                    cha_hidden,
                    kernel_size=5,
                    stride=1,
                    padding=2,
                    groups=cha_hidden,
                    bias=False,
                )
                self.conv4 = nn.utils.weight_norm(conv4, dim=None)

                self.avg_po_c4 = nn.AvgPool1d(kernel_size=4, stride=2, padding=1)

                self.flt = nn.Flatten()

                self.batch_norm2 = nn.BatchNorm1d(output_size)
                self.dropout2 = nn.Dropout(dropout_output)
                dense2 = nn.Linear(output_size, output_dim, bias=False)
                self.dense2 = nn.utils.weight_norm(dense2)
                self.silu = nn.SiLU()

                # self.loss = nn.BCEWithLogitsLoss()

            def forward(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                x = self.silu(x)

                return x

            def predict(self, x):
                x = self.batch_norm1(x)
                x = self.dropout1(x)
                x = nn.functional.celu(self.dense1(x))

                x = x.reshape(x.shape[0], self.cha_input, self.sign_size1)

                x = self.batch_norm_c1(x)
                x = nn.functional.relu(self.conv1(x))

                x = self.ave_po_c1(x)

                x = self.batch_norm_c2(x)
                x = self.dropout_c2(x)
                x = nn.functional.relu(self.conv2(x))
                x_s = x

                x = self.batch_norm_c3(x)
                x = self.dropout_c3(x)
                x = nn.functional.relu(self.conv3(x))

                x = self.batch_norm_c4(x)
                x = self.conv4(x)
                x = x + x_s
                x = nn.functional.relu(x)

                x = self.avg_po_c4(x)

                x = self.flt(x)

                x = self.batch_norm2(x)
                x = self.dropout2(x)
                x = self.dense2(x)

                x = self.silu(x)

                return x

        class Discriminator(nn.Module):
            """
            Will decide, if data is real or fake.
            """

            def __init__(self, num_features, dropout=0.3):
                super(Discriminator, self).__init__()

                self.dropout = dropout

                self.layer_0 = nn.Linear(num_features, 512)
                self.batch_norm_0 = nn.BatchNorm1d(512)

                self.layer_2 = nn.Linear(512, 16)
                self.batch_norm_2 = nn.BatchNorm1d(16)

                self.layer_3 = nn.Linear(16, 512)
                self.batch_norm_3 = nn.BatchNorm1d(512)

                self.layer_out = nn.Linear(512, 1)

                self.silu = nn.SiLU()

            def forward(self, inputs):
                x = self.silu(self.layer_0(inputs))
                x = self.batch_norm_0(x)
                x = self.silu(self.layer_2(x))
                x = self.batch_norm_2(x)
                x = self.silu(self.layer_3(x))
                x = self.batch_norm_3(x)
                x = self.layer_out(x)
                return x

            def predict(self, inputs):
                x = self.silu(self.layer_0(inputs))
                x = self.batch_norm_0(x)
                x = self.silu(self.layer_2(x))
                x = self.batch_norm_2(x)
                x = self.silu(self.layer_3(x))
                x = self.batch_norm_3(x)
                x = self.layer_out(x)
                return x

        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        n_features = X_train.values.shape[1]
        generator = Generator(
            num_features=n_features,
            output_dim=n_features,
            dropout_input=self.gan_settings["generator_dropout_rate"],
            dropout_hidden=self.gan_settings["generator_dropout_rate"],
            dropout_output=self.gan_settings["generator_dropout_rate"],
        )
        discriminator = Discriminator(num_features=n_features)

        generator.to(device)
        discriminator.to(device)
        return generator, discriminator

    def noise(self, n=None):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        n_features = X_train.values.shape[1]
        if not n:
            n = X_train.values.shape[0]
        return Variable(torch.randn(n, n_features)).to(device)

    def make_ones(self, size):
        data = Variable(torch.ones(size, 1))
        return data.to(device)

    def make_zeros(self, size):
        data = Variable(torch.zeros(size, 1))
        return data.to(device)

    def gan_model_setup(self):
        generator, discriminator = self.get_generator_discriminator()

        g_optim = optim.Adam(
            generator.parameters(),
            lr=self.gan_settings["generator_learning_rate"],
            betas=(0.5, 0.999),
        )
        d_optim = optim.Adam(
            discriminator.parameters(),
            lr=self.gan_settings["discriminator_learning_rate"],
            betas=(0.5, 0.999),
        )

        loss_fn = nn.BCEWithLogitsLoss()
        return generator, discriminator, g_optim, d_optim, loss_fn

    def boundary_seeking_loss(self, y_pred, y_true):
        return 0.5 * torch.mean((torch.log(y_pred) - torch.log(1 - y_pred)) ** 2)

    def train_discriminator(
        self, optimizer, real_data, fake_data, discriminator, loss_fn
    ):
        n = real_data.size(0)
        optimizer.zero_grad()
        real_pred = discriminator(real_data)
        fake_pred = discriminator(fake_data.detach())
        real_loss = loss_fn(real_pred - fake_pred, self.make_ones(n))
        fake_loss = loss_fn(fake_pred - real_pred, self.make_zeros(n))
        d_loss = (real_loss + fake_loss) / 2

        d_loss.backward()
        optimizer.step()

        return d_loss

    def train_generator(self, optimizer, fake_data, real_data, discriminator, loss_fn):
        n = fake_data.size(0)
        optimizer.zero_grad()

        discriminator(real_data).detach()
        discriminator(fake_data)
        g_loss = loss_fn(discriminator(fake_data), self.make_ones(n))

        g_loss.backward()
        optimizer.step()

        return g_loss, optimizer

    def create_gan_train_dataset(self, target_class):
        logging.info("Create NLP train dataset.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

        X_train_class_only = X_train.iloc[np.where(Y_train == target_class)[0]]
        Y_train_class_only = Y_train.iloc[np.where(Y_train == target_class)[0]]

        train_dataset = GanDataset(
            torch.from_numpy(X_train_class_only.to_numpy().astype(float)).float(),
            torch.from_numpy(Y_train_class_only.to_numpy().astype(float)).float(),
        )
        return train_dataset

    def create_gan_train_dataloader(
        self, target_class, train_batch_size=None, workers=None
    ):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.gan_settings["batch_size"]

        if workers:
            pass
        else:
            workers = self.gan_settings["num_workers"]
        logging.info("Create Neural network train dataloader.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        train_dataset = self.create_gan_train_dataset(target_class=target_class)
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=train_batch_size,
            shuffle=True,
            num_workers=workers,
            pin_memory=True,
        )
        return train_dataloader

    def train_gan(self, target_class):
        epochs = self.gan_settings["max_epochs"]
        k = self.gan_settings["discriminator_extra_training_rounds"]
        best_epoch = -1
        # test_noise = self.noise()
        train_dataloader = self.create_gan_train_dataloader(target_class=target_class)
        generator, discriminator, g_optim, d_optim, loss_fn = self.gan_model_setup()

        generator.train()
        discriminator.train()

        g_losses = []
        d_losses = []

        best_loss = 1000000000000
        for epoch in range(epochs):
            g_loss = 0.0
            d_loss = 0.0
            for i, data in enumerate(train_dataloader):
                imgs, _ = data
                n = len(imgs)
                for _j in range(k):
                    fake_data = generator(self.noise(n)).detach()
                    real_data = imgs.to(device)
                    d_loss += self.train_discriminator(
                        d_optim,
                        real_data,
                        fake_data,
                        discriminator=discriminator,
                        loss_fn=loss_fn,
                    )
                fake_data = generator(self.noise(n))
                real_data = imgs.to(device)
                g_loss_it, g_optim = self.train_generator(
                    g_optim,
                    fake_data,
                    real_data,
                    discriminator=discriminator,
                    loss_fn=loss_fn,
                )
                g_loss = g_loss + g_loss_it

                del fake_data
                del real_data
                _ = gc.collect()

                # img = generator(test_noise).cpu().detach()
                g_losses.append(g_loss / i)
                d_losses.append(d_loss / i)

            if epoch == 0 or g_loss < best_loss:
                print(
                    f"Found better model in epoch {epoch} with generator loss {g_loss}"
                    f" and discriminator loss {d_loss}."
                )
                best_epoch = epoch
                best_loss = g_loss
                state = {
                    "state_dict": generator.state_dict(),
                    "optimizer_dict": g_optim.state_dict(),
                    "bestscore": g_loss,
                }
                torch.save(state, f"class_{target_class}_generator_model.pth")
                # clear_output()

            if epoch > best_epoch + self.gan_settings["early_stopping_rounds"]:
                break

        del discriminator
        del generator
        torch.cuda.empty_cache()
        _ = gc.collect()

    def train_class_generators(self):
        self.get_current_timestamp(task="Start training GANs.")

        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        unique_classes = np.unique(Y_train.values)
        for clas in unique_classes:
            self.train_gan(target_class=clas)
        self.get_current_timestamp(task="Finished training GANs.")

    def load_generator_model_states(self, path=None, target_class=None):
        logging.info("Load model save states.")
        logging.info(f"RAM memory {psutil.virtual_memory()[2]} percent used.")
        if path:
            pass
        else:
            path = os.getcwd()
        if self.prediction_mode:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if re.search(f"class_{target_class}_generator_model.pth", s)
            ]
            return pthes
        else:
            pthes = [
                os.path.join(f"{path}/", s)
                for s in os.listdir(f"{path}/")
                if re.search(f"class_{target_class}_generator_model.pth", s)
            ]
            return pthes

    def create_synthetic_data(self):
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        columns = X_train.columns
        unique_classes = np.unique(Y_train.values)

        rows_to_create = self.gan_settings["nb_synthetic_rows_to_create"]
        batch_size = self.gan_settings["batch_size"]
        nb_batches = int(rows_to_create / batch_size)
        new_class_data = []
        new_targets = []

        for clas in unique_classes:
            new_data = []
            (
                generator,
                discriminator,
                g_optim,
                d_optim,
                loss_fn,
            ) = self.gan_model_setup()
            pathes = self.load_generator_model_states(
                path=self.gan_model_save_states_path, target_class=clas
            )
            print(pathes)
            for m_path in pathes:
                state = torch.load(m_path)
                generator.load_state_dict(state["state_dict"])
                generator.to(device)
                generator.eval()
                with torch.no_grad():
                    for _batch in range(nb_batches):
                        synth_data = generator.predict(self.noise(batch_size))
                        synth_data = synth_data.cpu().detach()
                        new_data.append(synth_data)

            full_new_data = np.concatenate(new_data, axis=0)
            new_train = pd.DataFrame(full_new_data, columns=columns)
            new_train[self.target_variable] = clas

            new_target = new_train[self.target_variable]
            new_targets.append(new_target.values)

            new_train = new_train.drop(self.target_variable, axis=1)
            new_class_data.append(new_train)

            del generator

        if self.gan_settings["concat_to_original_data"]:
            X_train[self.target_variable] = Y_train
            X_train = X_train.sort_index(axis=1)
            print(X_train.head(3))
            original_columns = X_train.columns
            X_train_synth = pd.concat(new_class_data)
            Y_train_synth = pd.Series(np.concatenate(new_targets, axis=0))
            print(X_train_synth.head(3))
            X_train_synth[self.target_variable] = Y_train_synth.values
            X_train = pd.concat(
                [X_train[original_columns], X_train_synth[original_columns]]
            )
        else:
            X_train = pd.concat(new_class_data)
            Y_train = pd.Series(np.concatenate(new_targets, axis=0))
            X_train[self.target_variable] = Y_train.values

        if len(X_train.index) > rows_to_create:
            X_train = X_train.sample(
                rows_to_create, random_state=self.global_random_state
            )

        X_train = X_train.reset_index(drop=True)
        Y_train = X_train[self.target_variable].copy()
        X_train = X_train.drop(self.target_variable, axis=1)

        try:
            X_test = X_test.drop(self.target_variable, axis=1)
        except KeyError:
            pass

        self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)
